/*
 * draw.c
 * Copyright (C) Steffen Pohle 2010 <steffen@gulpe.de>
 * 
 * main.c is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * main.c is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "osmroute.h"
#include "draw.h"
#include "map.h"
#include "gps.h"
#include "system.h"

#include <sys/time.h>


/**************************************************************************
 * color definitions, hopfully this can go out soon, need to find a way to 
 * specify only the rgb values like it is in HTML and thats it.
 * i.e. color = 0xFF0000; for red.
 */
struct color *draw_btn[DRAW_BTN_MAX];
struct line_style draw_linestyle[MWAY_MAX];
struct color color_polygon[MAREA_MAX];
struct color color_bg;
struct color color_fg;
struct color color_white, color_grey, color_red1, color_red2, color_blue1, color_blue2, color_green1, color_green2, color_pink, color_black;


/**************************************************************************
 * current screen data, center of the screen? where we look at the scale,
 * and the drawing flags. All these values start with view_*
 */
float view_scale_list[] = {0.000675, 0.00125, 0.0025, 0.005, 0.01, 0.025, 0.05, 0.1, 0.25, 0.5, 1.0, 0.0};
int view_scale = 3;
float view_lon = 11.183;
float view_lat = 48.736;
int view_flags = DRAW_GPSFOLLOW;


/**************************************************************************
 * internal variables set by some functions within the drawing stuff and
 * from withing the *_gfx sources.. Some drawing functon will have vriables 
 * declared on top of the first function. (i.E. drawing lines, polygones.
 */
int draw_type[MAREA_MAX];
iPoint draw_screensize = { 0, 0 };
struct map_pos draw_pointer = {0.0, 0.0};
int draw_pointer_mode = 0;					// needed for pressing modes (long, short)
struct gps_data draw_gpsdata;


iPoint draw_geo2screen (float center_lon, float center_lat, float lon, float lat);
void draw_screen2geo (int x, int y, float *lon, float *lat);

void draw_line (struct image *img, int x1, int y1, int x2, int y2, struct line_style style);
void draw_lines (struct image *img);
void draw_lineway (struct image *img, int x1, int y1, int x2, int y2, struct line_style style);
void draw_rectangle (struct image *img, int x1, int y1, int x2, int y2, struct line_style style);
void draw_polygonstart ();
void draw_polygonadd (int x, int y);
void draw_polygonfinish (struct image *img, struct line_style ls, struct color c);


/*************************************************************************
 * to prevent the flickering screen, we create two images inside the 
 * memory. One which keeps a larger area of the map (so we only draw 
 * the map once in a while) and a second map with the on screen data.
 * img_map: keeps the current map screen values.
 * img_screen: is the final data with the on screen data to copy on
 * the screen.
 */
struct image *img_map = NULL;		// image data of the current map
struct map_pos img_map_pos;			// position of the current map
int img_map_scale;					// scale of the current map
struct image *img_screen = NULL;	// screen data needed for double buffering



/**************************************************************************
 *
 * functions to convert between geo data and screen data
 *
 */
iPoint draw_geo2screen (float center_lon, float center_lat, float lon, float lat) {
	iPoint r;
	
	r.x = - (map_lon2km (center_lon - lon, lat) / view_scale_list[view_scale]);
	r.y = (map_lat2km (center_lat - lat) / view_scale_list[view_scale]);

	return r;
};


void draw_screen2geo (int x, int y, float *lon, float *lat) {
	int sx, sy;

	draw_getsize (&sx, &sy);
	y = y - sy/2;
	x = x - sx/2;
			
	*lat = -map_km2lat(((float)y) * view_scale_list[view_scale]);
	*lon = map_km2lon(((float)x) * view_scale_list[view_scale], view_lat + *lat);
	*lat = view_lat + *lat;
	*lon = view_lon + *lon;
};


/******************************************************************************
 * set the new scale, and check what will have to be shown
 */
void draw_setscale (int newscale) {
	int i;

	if (newscale == view_scale) return;
	
	for (i = 0; view_scale_list[i] > 0.0; i++);	/* find max i in view_scale_list */

	if (newscale < 0) view_scale = 0;
	else if (newscale >= i) view_scale = i-1;
	else view_scale = newscale;

	if (view_scale_list[view_scale] < 0.01) {
		for (i = 0; i < MWAY_MAX; i++)
			draw_type [i] = TRUE;
	}
	else if (view_scale_list[view_scale] < 0.1) {
		for (i = 0; i < MWAY_MAX; i++) {
			if (i <= MWAY_track) draw_type [i] = TRUE;
			else  draw_type [i] = FALSE;
		}
	}
	else {
		for (i = 0; i < MWAY_MAX; i++) {
			if (i == MWAY_highway) draw_type [i] = TRUE;
			else  draw_type [i] = FALSE;
		}
	}

	main_wnd_update();
};


void draw_set_pointerpos (float lon, float lat, int mode) {
	if (mode != 0) draw_pointer_mode = mode;
	else draw_pointer_mode = 0;

	draw_pointer.lon = lon;
	draw_pointer.lat = lat;
};


void draw_set_gpspos (struct gps_data *gpsdata) {
	static char _gpspos_to = 0;
	float lon, lat;
	
	if (gpsdata->valid) {
		memcpy (&draw_gpsdata, gpsdata, sizeof (struct gps_data));
		draw_set_flag (DRAW_GPS);
		
		if (view_flags & DRAW_GPSFOLLOW && buttonpressed == 0) {
			iPoint p = 	draw_geo2screen (view_lon, view_lat, draw_gpsdata.lon, draw_gpsdata.lat);
			if (p.x < (draw_screensize.x / 4) || p.x > (draw_screensize.x - (draw_screensize.x/4)) ||
			   (p.y < (draw_screensize.y / 4) || p.y > (draw_screensize.y - (draw_screensize.y/4)))) {
					view_lon = draw_gpsdata.lon;
					view_lat = draw_gpsdata.lat;
				}
		}

		/* find closest way */
		selected_way_pos[0].lon = draw_gpsdata.lon;
		selected_way_pos[0].lat = draw_gpsdata.lat;
		selected_way = map_way_findclosest (&selected_way_pos[0], &selected_way_pos[1], 0, selected_way_name, MAP_W_NAMELEN);
	}
	if (gpsdata->quality != 0 || (view_flags & DRAW_DEBUG) && ((_gpspos_to--) < 0)) {
		draw ();
		_gpspos_to = 30;
	}
};


void draw_set_flag (int flag) {
	view_flags |= flag;
};


void draw_del_flag (int flag) {
	view_flags &= (0xFFFF - flag);
};


/*****************************************************************
 * returns DRAW_BTN_* for the button which is under the mouse
 * BUTTON POS: sx = 
 */
int draw_button_check (int x, int y) {
	int bx;
	int i;
	int sx;
	int sy;
	int btn;

	/* init the values */
	bx = draw_screensize.x/DRAW_BTN_MAX;
	if (bx == 0) return DRAW_BTN_NONE;
	i = x / bx;
	sx = i * bx + (bx-DRAW_BTN_WIDTH)/2;
	sy = draw_screensize.y-DRAW_BTN_WIDTH;
	btn = DRAW_BTN_NONE;
	
	
	if (y >= sy && y <= sy+DRAW_BTN_WIDTH && x >= sx && x <= sx+DRAW_BTN_WIDTH) btn = i+1;
	// d_printf ("draw_button_check sx:%d sy:%d i:%d x:%d y:%d, btn:%d", sx, sy, i, x, y, btn);

	if (btn == DRAW_BTN_GPS) {
		if (gps_isrunning () <= 0) {
			gps_start ();
			main_wnd_update();
		}
		else {
			if (view_flags & DRAW_GPSFOLLOW) {
				draw_del_flag (DRAW_GPSFOLLOW);
				draw_btn[DRAW_BTN_GPS-1] = &color_green1;
			}
			else {
				draw_set_flag (DRAW_GPSFOLLOW);
				draw_btn[DRAW_BTN_GPS-1] = &color_green2;
			}
		}
	}
	else if (btn == DRAW_BTN_SCALEINC) draw_setscale (view_scale - 1);
	else if (btn == DRAW_BTN_SCALEDEC) draw_setscale (view_scale + 1);
	else if (btn == DRAW_BTN_DEBUG && (view_flags & DRAW_DEBUG)) {
		draw_del_flag (DRAW_DEBUG);
		draw_btn[DRAW_BTN_DEBUG-1] = &color_grey;
	}
	else if (btn == DRAW_BTN_DEBUG && !(view_flags & DRAW_DEBUG)) {
		draw_set_flag (DRAW_DEBUG);
		draw_btn[DRAW_BTN_DEBUG-1] = &color_red2;
	}	
	if (btn != DRAW_BTN_NONE) draw ();
	
	return btn;
};


/*
 * sets the color
 */
void draw_button_set (int btn, struct color *c) {
	if (btn > DRAW_BTN_MAX || btn <= 0) {
		d_printf ("draw_button_set btn=%d DRAW_BTN_MAX=%d", btn, DRAW_BTN_MAX);
		return;
	}

	draw_btn[btn-1] = c;
};


/**************************************************************************+
 * drawing subfunctions
 */

/*
 * rectangle
 */
void draw_rectangle (struct image *img, int x1, int y1, int x2, int y2, struct line_style style) {
	gfx_img_draw_line (img, x1, y1, x2, y1, style);
	gfx_img_draw_line (img, x2, y1, x2, y2, style);
	gfx_img_draw_line (img, x2, y2, x1, y2, style);
	gfx_img_draw_line (img, x1, y2, x1, y1, style);
};


/*
 * lines
 */
static int _linescnt;
static iPoint _linesp1[LINES_MAX];
static iPoint _linesp2[LINES_MAX];
static struct line_style _linesstyle[LINES_MAX];
static int _lineswarning = 0;

void draw_line (struct image *img, int x1, int y1, int x2, int y2, struct line_style style) {
	if ((x1 >= 0 && x1 <= img->width && y1 >= 0 && y1 <= img->height) ||
        (x2 >= 0 && x2 <= img->width && y2 >= 0 && y2 <= img->height)) {       	
			if (_linescnt >= LINES_MAX) {
				if (_lineswarning == 0) {
					printf ("%s:%d draw_line: too many lines.\n", __FILE__, __LINE__);
					_lineswarning = 1;
				}
				return;
			}

			_linesp1[_linescnt].x = x1;	
			_linesp1[_linescnt].y = y1;
			_linesp2[_linescnt].x = x2;
			_linesp2[_linescnt].y = y2;
			_linesstyle[_linescnt] = style;
			_linescnt++;
		}
};


void draw_lines (struct image *img) {
	int i;
	float j;
	float ox;
	float oy;
	float angle;
	
	/* draw the outside */
	for (i = 0; i < _linescnt; i++) {
		gfx_img_draw_line (img, _linesp1[i].x, _linesp1[i].y, _linesp2[i].x, _linesp2[i].y, _linesstyle[i]);
	}

	_linescnt = 0;
};


/*
 * draw a polygon..
 */
static iPoint _polypoints[POLYGON_MAX_POINTS];
static int _polycnt;

void draw_polygonstart () {
	_polycnt = 0;
};

void draw_polygonadd (int x, int y) {
	if (_polycnt >= POLYGON_MAX_POINTS) {
		printf ("%s:%d draw_polygonadd: too many points.\n", __FILE__, __LINE__);
		return;
	}
	_polypoints[_polycnt].x = x;
	_polypoints[_polycnt].y = y;
	_polycnt++;
};


void draw_polygonfinish (struct image *img, struct line_style ls, struct color c) {
	gfx_img_draw_polygon (img, _polypoints, _polycnt, ls, c);
	_polycnt = 0;
};


void draw_lineway (struct image *img, int x1, int y1, int x2, int y2, struct line_style style) {
	struct line_style tmplinestyle = style;

	if (view_scale_list[view_scale] < 0.01) 
		tmplinestyle.width = tmplinestyle.width * (0.005/view_scale_list[view_scale]);

	draw_line (img, x1, y1, x2, y2, tmplinestyle);
};


/************************************************************************
 * the drawing stuff itself.
 * before drawing everything on the screen, check if we need to redraw 
 * the map or just copy a part of it into the gui image.
 * After copying the map into the gui screen draw the gui, afterwards 
 * copy the gui screen onto the display.
 */

void draw_map () {
	struct map_hash *mh, *tmpmh;
	struct map_way *tmpway, *cway;
	struct map_area *carea;
	struct map_place *cplace;
	int w, n, i;
	struct map_hashpos h, hs, he;
	iPoint p1, p2, p3, p4;
	fPoint nd;
	unsigned long long int wid;
	struct line_style tmplinestyle;
	int drawgrid = 0;
	char text[255];
	iPoint center;
	struct map_pos hashstart, hashend, tmppos;

	gfx_img_clear (img_map, &color_bg);

	center.x = img_map->width / 2;
	center.y = img_map->height / 2;

	tmplinestyle.width = 1.0;
	tmplinestyle.c = tmplinestyle.borderc = color_red1;
	
//	p1.y = p1.x = 0;
//	p2 = p1;

	hashstart.lat = view_lat - map_km2lat(view_scale_list[view_scale] * ((float)img_map->width));
	hashstart.lon = view_lon - map_km2lon(view_scale_list[view_scale] * ((float)img_map->height), hashstart.lat);
	hashend.lat = view_lat + map_km2lat(view_scale_list[view_scale] * ((float)img_map->height));
	hashend.lon = view_lon + map_km2lon(view_scale_list[view_scale] * ((float)img_map->width), hashend.lat);

	hs.ilon = map_geo2igeo (hashstart.lon);
	hs.ilat = map_geo2igeo (hashstart.lat);
	he.ilon = map_geo2igeo (hashend.lon);
	he.ilat = map_geo2igeo (hashend.lat);

	/*
	 * draw the lon / lat grid
	 */
	for (h.ilon = hs.ilon; h.ilon <= he.ilon; h.ilon++) 
		for (h.ilat = hs.ilat; h.ilat <= he.ilat; h.ilat++) {
			p1 = draw_geo2screen (view_lon, view_lat, map_igeo2geo(h.ilon), map_igeo2geo(h.ilat));
			p2 = draw_geo2screen (view_lon, view_lat, map_igeo2geo(h.ilon+1), map_igeo2geo(h.ilat));
			p3 = draw_geo2screen (view_lon, view_lat, map_igeo2geo(h.ilon), map_igeo2geo(h.ilat+1));

			draw_line (img_map, p1.x, p1.y, p2.x, p2.y, tmplinestyle);
			draw_line (img_map, p1.x, p1.y, p3.x, p3.y, tmplinestyle);
		}

/*	for (h.ilon = hs.ilon; h.ilon <= he.ilon; h.ilon++) 
		for (h.ilat = hs.ilat; h.ilat <= he.ilat; h.ilat++) {
			if (h.ilon % MAP_LSHASHFACTOR == 0 && h.ilat % MAP_LSHASHFACTOR == 0) {
				tmplinestyle.width = 1.0;
				tmplinestyle.c = tmplinestyle.borderc = color_red2;
				p1 = draw_getposlonlat (map_igeo2geo(h.ilon), map_igeo2geo(h.ilat));
				p2 = draw_getposlonlat (map_igeo2geo(h.ilon+1*MAP_LSHASHFACTOR), map_igeo2geo(h.ilat));
				p3 = draw_getposlonlat (map_igeo2geo(h.ilon), map_igeo2geo(h.ilat+1*MAP_LSHASHFACTOR));
				p4 = draw_getposlonlat (map_igeo2geo(h.ilon+1*MAP_LSHASHFACTOR), map_igeo2geo(h.ilat+1*MAP_LSHASHFACTOR));
				draw_line (p1.x, p1.y, p2.x, p2.y, tmplinestyle);
				draw_line (p1.x, p1.y, p3.x, p3.y, tmplinestyle);
			}
		}
*/
	/*
	 * draw areas
	 */
	for (h.ilon = hs.ilon; h.ilon <= he.ilon; h.ilon++) 
		for (h.ilat = hs.ilat; h.ilat <= he.ilat; h.ilat++) {
			if (app_status == APPSTATUS_nothing) mh = map_hash_geti (h.ilon, h.ilat, MHLOAD_RELOAD);
			else mh = map_hash_geti (h.ilon, h.ilat, MHLOAD_NONE);
			if (mh && view_scale_list[view_scale] < 0.1) {
				carea = mh->areas;
				while (carea) {
					draw_polygonstart ();	
					for (n = 0; n < carea->p_cnt; n++) {
						p1 = draw_geo2screen (view_lon, view_lat, carea->p[n].lon, carea->p[n].lat);
						draw_polygonadd (p1.x + center.x, p1.y + center.y);
					}
					draw_polygonfinish (img_map, tmplinestyle, color_polygon[carea->type]);
					carea = carea->next;
				}
			}
		}


	/*
	 * draw ways
	 */
	for (h.ilon = hs.ilon; h.ilon <= he.ilon; h.ilon++) 
		for (h.ilat = hs.ilat; h.ilat <= he.ilat; h.ilat++) {
			/*
			 * draw the hash.. 
			 */
			if (app_status == APPSTATUS_nothing) mh = map_hash_geti (h.ilon, h.ilat, MHLOAD_RELOAD);
			else  mh = map_hash_geti (h.ilon, h.ilat, MHLOAD_NONE);
			if (mh) {
				/*
				 * ways
				 */
				cway = mh->ways;
				while (cway) {
					if (cway->id != 0 && draw_type[cway->type]) {
						/* go through all the nodes */
						for (n = 0; n < cway->p_cnt; n++) {
							p2 = p1;
							nd.x = map_lon2km (view_lon - cway->p[n].lon, cway->p[n].lat);
							nd.y = map_lat2km (view_lat - cway->p[n].lat);
							p1.x = center.x - (nd.x / view_scale_list[view_scale]);
							p1.y = center.y + (nd.y / view_scale_list[view_scale]);
					
							if (n > 0) {
	//							if (cway->id == selected_way) {
	//								tmplinestyle.width = 4.0;
	//								tmplinestyle.c = tmplinestyle.borderc = color_blue2;
	//								draw_lineway (img_map, p1.x, p1.y, p2.x, p2.y, tmplinestyle);
	//							}
								if (cway->type >= MWAY_MAX)
									draw_lineway (img_map, p1.x, p1.y, p2.x, p2.y, draw_linestyle[MWAY_unknown]);
								else
									draw_lineway (img_map, p1.x, p1.y, p2.x, p2.y, draw_linestyle[cway->type]);
							}
						}

						/* draw the waynodes */
						for (n = 0; n < cway->n_cnt; n++) if (cway->n[n].pnr < cway->p_cnt) {
							p1 = draw_geo2screen (view_lon, view_lat, cway->n[n].d.lon, cway->n[n].d.lat);
							p2 = draw_geo2screen (view_lon, view_lat, cway->p[cway->n[n].pnr].lon, cway->p[cway->n[n].pnr].lat);
							p1.x += center.x;
							p1.y += center.y;
							p2.x += center.x;
							p2.y += center.y;
							if (cway->type >= MWAY_MAX)
								draw_lineway (img_map, p1.x, p1.y, p2.x, p2.y, draw_linestyle[MWAY_unknown]);
							else
								draw_lineway (img_map, p1.x, p1.y, p2.x, p2.y, draw_linestyle[cway->type]);
						}
						else d_printf ("%s:%d wayid:%lld cway->n[n].pnr(%d) < cway->p_cnt(%d) doesn't match", __FILE__, __LINE__, cway->id, cway->n[n].pnr, cway->p_cnt);
	
					}
					cway = cway->next;
				}
			}
		}
	
	draw_lines (img_map);
	/*
	 * draw places, pois just on top layers
	 */
	for (h.ilon = hs.ilon; h.ilon <= he.ilon; h.ilon++) 
		for (h.ilat = hs.ilat; h.ilat <= he.ilat; h.ilat++) {

			if (app_status == APPSTATUS_nothing) mh = map_hash_geti (h.ilon, h.ilat, MHLOAD_RELOAD);
			else mh = map_hash_geti (h.ilon, h.ilat, MHLOAD_NONE);
			if (mh) {
				/*
				 * places
				 */
				cplace = mh->places;
				i = 0;
				while (cplace) {
					if (view_scale_list[view_scale] < 0.01 || cplace->type == MAP_PLACE_city) {
						i++;
						p1 = draw_geo2screen(view_lon, view_lat, cplace->lon, cplace->lat);
						p1.x += center.x;
						p1.y += center.y;
						switch (cplace->type) {
							case (MAP_PLACE_city):
							case (MAP_PLACE_town):
							case (MAP_PLACE_village):
								gfx_img_draw_text (img_map, p1.x,p1.y, cplace->name, &color_white);
								break;
							case (MAP_PLACE_hamlet):
							case (MAP_PLACE_suburb):
								gfx_img_draw_text (img_map, p1.x,p1.y, cplace->name, &color_blue1);
								break;
							default:
								gfx_img_draw_text (img_map, p1.x,p1.y, cplace->name, &color_red2);
								break;
						}
					}
					cplace = cplace->next;
				}
			}
		}

	// tmplinestyle.width = 4.0;
	// tmplinestyle.c = tmplinestyle.borderc = color_red1;
	// draw_line (img_map, center.x-100, center.y, center.x+100, center.y, tmplinestyle);
	// draw_line (img_map, center.x, center.y-100, center.x, center.y+100, tmplinestyle);
	// draw_lines (img_map);
	// sprintf (text, "%f,%f", view_lon, view_lat);
	// gfx_img_draw_text (img_map, center.x,center.y, text, &color_blue1);

	/*
	 * save data of the map drawed
	 */
	img_map_pos.lon = view_lon;
	img_map_pos.lat = view_lat;
	img_map_scale = view_scale;
};


int draw_mapcopy () {
	float x, y;
	iPoint p1, p2;

	p2 = draw_geo2screen (img_map_pos.lon, img_map_pos.lat, view_lon, view_lat);
	p1.x = img_map->width/2 + p2.x - img_screen->width/2;
	p1.y = img_map->height/2 + p2.y - img_screen->height/2;

	if (p1.x < 0 || p1.y < 0 
	    || p1.x > (img_map->width - img_screen->width) 
	    || p1.y > (img_map->height - img_screen->height)) return 0;
	
	gfx_img_draw_img (img_screen, 0, 0, img_screen->width, img_screen->height, img_map, p1.x, p1.y);

	return 1;
};


void draw_gui () {
	struct timeval time_start, time_end;
	struct line_style tmplinestyle;
	iPoint p1, p2, center;
	int i;
	char *c;
	char text[64];
	struct gps_data *gpsdata;

	center.x = draw_screensize.x / 2;
	center.y = draw_screensize.y / 2;

	/*
	 * draw selected line
	 */
	tmplinestyle.width = 1.0;
	tmplinestyle.c = tmplinestyle.borderc = color_red2;
	if (selected_way != 0) {
		p1 = draw_geo2screen (view_lon, view_lat, selected_way_pos[0].lon, selected_way_pos[0].lat);
		p1.x += center.x;
		p1.y += center.y;
		p2 = draw_geo2screen (view_lon, view_lat, selected_way_pos[1].lon, selected_way_pos[1].lat);
		p2.x += center.x;
		p2.y += center.y;
		draw_line (img_screen, p1.x, p1.y, p2.x, p2.y, tmplinestyle);
	}
	
	
	/*
	 * gpsroute 
	 */
	if (gpsroute) {
		if (gpsroute->cnt > 1) {
			tmplinestyle.width = 1.0;
			tmplinestyle.c = tmplinestyle.borderc = color_red2;

			p1 = draw_geo2screen (view_lon, view_lat, gpsroute->data[0].lon, gpsroute->data[0].lat);
			p1.x += center.x;
			p1.y += center.y;
			for (i = 1; i < gpsroute->cnt; i++) {
				p2 = draw_geo2screen (view_lon, view_lat, gpsroute->data[i].lon, gpsroute->data[i].lat);
				p2.x += center.x;
				p2.y += center.y;
				if ((p1.x >= 0 && p1.x < draw_screensize.x && p1.y >= 0 && p1.y < draw_screensize.y) || 
				    (p2.x >= 0 && p2.x < draw_screensize.x && p2.y >= 0 && p2.y < draw_screensize.y))
					draw_line (img_screen, p1.x, p1.y, p2.x, p2.y, tmplinestyle);
				p1 = p2;
			}
		}
		
		if (gpsroute->pcnt > 0) {
			tmplinestyle.width = 1.0;
			tmplinestyle.c = tmplinestyle.borderc = color_red1;

			for (i = 0; i < gpsroute->pcnt; i++) {
				p1 = draw_geo2screen (view_lon, view_lat, gpsroute->pdata[i].lon, gpsroute->pdata[i].lat);
				p1.x += center.x;
				p1.y += center.y;
				draw_line (img_screen, p1.x-5, p1.y-5, p1.x+5, p1.y+5, tmplinestyle);
				draw_line (img_screen, p1.x+5, p1.y-5, p1.x-5, p1.y+5, tmplinestyle);
				gfx_img_draw_text (img_screen, p1.x-10, p1.y+5, gpsroute->pdata[i].name, &color_red2);
			}
		}
	}

	
	/*
	 * draw gps position
	 */
	if (view_flags & DRAW_GPS) {
		iPoint p3, p4, p;
		p = draw_geo2screen (view_lon, view_lat, draw_gpsdata.lon, draw_gpsdata.lat);
		p.x += center.x;
		p.y += center.y;

		tmplinestyle.width = 1.0;
		tmplinestyle.c = tmplinestyle.borderc = color_blue2;
		for (i = 1; i < 10; i++) {
			p1.x = sin (M_PI*draw_gpsdata.direction/180.0) * 12 + p.x;
			p1.y = -cos (M_PI*draw_gpsdata.direction/180.0) * 12 + p.y;

			p2.x = sin (M_PI*(draw_gpsdata.direction-120.0)/180.0) * i + p.x;
			p2.y = -cos (M_PI*(draw_gpsdata.direction-120.0)/180.0) * i + p.y;

			p3.x = sin (M_PI*(draw_gpsdata.direction+120.0)/180.0) * i + p.x;
			p3.y = -cos (M_PI*(draw_gpsdata.direction+120.0)/180.0) * i + p.y;

			draw_line (img_screen, p1.x, p1.y, p2.x, p2.y, tmplinestyle);
			draw_line (img_screen, p1.x, p1.y, p3.x, p3.y, tmplinestyle);
			draw_line (img_screen, p3.x, p3.y, p2.x, p2.y, tmplinestyle);
		}
		i = 6;
		p1.x = sin (M_PI*draw_gpsdata.direction/180.0) * 8 + p.x;
		p1.y = -cos (M_PI*draw_gpsdata.direction/180.0) * 8 + p.y;

		p2.x = sin (M_PI*(draw_gpsdata.direction-120.0)/180.0) * 5 + p.x;
		p2.y = -cos (M_PI*(draw_gpsdata.direction-120.0)/180.0) * 5 + p.y;

		p3.x = sin (M_PI*(draw_gpsdata.direction+120.0)/180.0) * 5 + p.x;
		p3.y = -cos (M_PI*(draw_gpsdata.direction+120.0)/180.0) * 5 + p.y;
	
		tmplinestyle.width = 1.0;
		tmplinestyle.c = tmplinestyle.borderc = color_white;
		draw_line (img_screen, p1.x, p1.y, p2.x, p2.y, tmplinestyle);
		draw_line (img_screen, p1.x, p1.y, p3.x, p3.y, tmplinestyle);
		draw_line (img_screen, p2.x, p2.y, p3.x, p3.y, tmplinestyle);
	}

	/*
	 * our mouse cursor 
	 */
	tmplinestyle.width = 2.0;
	if (draw_pointer_mode == 1)
		tmplinestyle.c = tmplinestyle.borderc = color_red2;
	else
		tmplinestyle.c = tmplinestyle.borderc = color_white;
	p1 = draw_geo2screen (view_lon, view_lat, draw_pointer.lon, draw_pointer.lat);
	p1.x += center.x;
	p1.y += center.y;
	
	draw_line (img_screen, p1.x-5, p1.y-5, p1.x+5, p1.y+5, tmplinestyle);
	draw_line (img_screen, p1.x+5, p1.y-5, p1.x-5, p1.y+5, tmplinestyle);

	if (view_flags & DRAW_DEBUG) {
		sprintf (text, "%d,%d", map_geo2igeo (draw_pointer.lon), map_geo2igeo (draw_pointer.lat));
		gfx_img_draw_text (img_screen, p1.x-30, p1.y-16, text, &color_white);
		sprintf (text, "%d,%d", map_geo2igeo (draw_pointer.lon)/MAP_LSHASHFACTOR, map_geo2igeo (draw_pointer.lat)/MAP_LSHASHFACTOR);
		gfx_img_draw_text (img_screen, p1.x-20, p1.y+5, text, &color_white);
	}

	// tmplinestyle.width = 1.0;
	// tmplinestyle.c = tmplinestyle.borderc = color_red2;
	// draw_line (0, 0, screensize.x-1, screensize.y-1, tmplinestyle);
	// draw_line (screensize.x-1, 0, 0, screensize.y-1, tmplinestyle);
	
	if (view_flags & DRAW_DEBUG) {
		
		if (gps_isrunning()) {
			gpsdata = gps_getposition ();
			/* draw sat satus */
			tmplinestyle.width = 1.0;
			tmplinestyle.c = tmplinestyle.borderc = color_grey;
			p1.x = 20;
			p1.y = 60;
			p2.x = 20 + 99;
			p2.y = 60 + 32 * 3;
			for (i = 0; i < 2; i++) 
				draw_line (img_screen, p1.x+i*25, p1.y, p1.x+i*25, p2.y, tmplinestyle);
			tmplinestyle.c = tmplinestyle.borderc = color_white;
			for (i = 0; i < 32; i++) {
				p1.x = 20;
				p1.y = 60 + i * 3;
				p2.x = 20 + gpsdata->sat[i].quality;
				p2.y = 60 + i * 3;
				draw_line (img_screen, p1.x, p1.y, p2.x, p2.y, tmplinestyle);
			}

			sprintf (text, "Lon:%7.3f Lat:%7.3f Height:%4.0f valid:%d", gpsdata->lon, gpsdata->lat, gpsdata->height, gpsdata->valid);
			gfx_img_draw_text (img_screen, 0, 5, text, &color_green2);
			sprintf (text, "Sats:%d Quality:%d HDOP:%5.2f", gpsdata->numofsatelites, gpsdata->quality, gpsdata->hdop);
			gfx_img_draw_text (img_screen, 0, 20, text, &color_green2);
		}
	}

	if (view_flags & DRAW_GPS) {
		sprintf (text, "%3.0f km/h :%s", draw_gpsdata.speed, selected_way_name);
		gfx_img_draw_text (img_screen, 0, draw_screensize.y-32-16, text, &color_green2);
	}

	
	draw_lines (img_screen);
	draw_buttons ();
};


int draw_buttons () {
	int bx = draw_screensize.x/DRAW_BTN_MAX;
	int i;
	int sx;
	int sy = draw_screensize.y-DRAW_BTN_WIDTH;
	struct line_style tmplinestyle;

	tmplinestyle.width = 1.0;
	tmplinestyle.c = tmplinestyle.borderc = color_white;
	
	for (i = 0; i < DRAW_BTN_MAX; i++) {
		sx = i * bx + (bx-DRAW_BTN_WIDTH)/2;
		draw_polygonstart ();
		draw_polygonadd (sx, sy);
		draw_polygonadd (sx+DRAW_BTN_WIDTH, sy);
		draw_polygonadd (sx+DRAW_BTN_WIDTH, sy+DRAW_BTN_WIDTH);
		draw_polygonadd (sx, sy+DRAW_BTN_WIDTH);
		draw_polygonfinish (img_screen, tmplinestyle, *draw_btn[i]);

		if (i == DRAW_BTN_SCALEINC-1) gfx_img_draw_text (img_screen, sx+2, sy+2, "+", &color_white);
		if (i == DRAW_BTN_SCALEDEC-1) gfx_img_draw_text (img_screen, sx+2, sy+2, "-", &color_white);
		if (i == DRAW_BTN_GPS-1) gfx_img_draw_text (img_screen, sx+2, sy+2, "G", &color_white);
		if (i == DRAW_BTN_DEBUG-1 && (view_flags & DRAW_DEBUG)) gfx_img_draw_text (img_screen, sx+2, sy+2, "D", &color_white);
		if (i == DRAW_BTN_DEBUG-1 && !(view_flags & DRAW_DEBUG)) gfx_img_draw_text (img_screen, sx+2, sy+2, "D", &color_white);

		draw_line (img_screen, sx, sy, sx+DRAW_BTN_WIDTH, sy, tmplinestyle);
		draw_line (img_screen, sx+DRAW_BTN_WIDTH, sy, sx+DRAW_BTN_WIDTH, sy+DRAW_BTN_WIDTH, tmplinestyle);
		draw_line (img_screen, sx+DRAW_BTN_WIDTH, sy+DRAW_BTN_WIDTH, sx, sy+DRAW_BTN_WIDTH, tmplinestyle);
		draw_line (img_screen, sx, sy+DRAW_BTN_WIDTH, sx, sy, tmplinestyle);
	}
	draw_lines (img_screen);
};


void draw () {
	if (!draw_start ())	return;

	draw_getsize (&draw_screensize.x, &draw_screensize.y);

	/* do we need to alloc the screens */
	if (img_map != NULL && (img_map->width != draw_screensize.x * 2 || img_map->height != draw_screensize.y * 2)) {
		gfx_img_free (img_map);
		img_map = NULL;
	}
	if (img_map == NULL) {
		img_map = gfx_img_alloc (draw_screensize.x * 2, draw_screensize.y * 2);
		img_map_scale = -1;
	}

	if (img_screen != NULL && (img_screen->width != draw_screensize.x || img_screen->height != draw_screensize.y)) {
		gfx_img_free (img_screen);
		img_screen = NULL;
	}
	if (img_screen == NULL) {
		img_screen = gfx_img_alloc (draw_screensize.x, draw_screensize.y);
	}

	/* draw the map, if needed */
	if (img_map_scale < 0 || img_map_scale != view_scale) {
		draw_map ();
	}
	
	if (draw_mapcopy () == 0) {
		draw_map ();
		draw_mapcopy ();
	}
	
	draw_gui ();
	gfx_draw_img (0, 0, draw_screensize.x, draw_screensize.y, img_screen, 0, 0);

	draw_stop ();
};


void draw_redrawmap () {
	img_map_scale = -1;
	draw ();
};
