/***************************************************************************
 *            map_osmload.c
 *
 *  2009-08-16
 *  Copyright  2009  Steffen Pohle
 *  steffen@gulpe.de
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/***************************************************
 *
 * read the xml file provided by OpenStreetMap
 *
 */

#define _FILE_OFFSET_BITS 64

#include <locale.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <dirent.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
// #include <sys/wait.h>
#include <libxml/xmlreader.h>
#include <bzlib.h>

#include "osmroute.h"
#include "map.h"
#include "system.h"
#include "memoryleak.h"

#define TMPDIR "/tmp/gosmroute"

struct s_osm_node {
	char *mem;
	int memsize;
	char hid;
	int fd;
	int fsize;
	int fpos;
	int mpos;
	int changed;
};

struct s_osm_hashnodes {
	short int hid;	// number of the hid, -1 == unused

	int hash_cnt;	// current numbers inside
	int hash_fd;	// file descriptor
	int hash_alloc;
	struct osm_nodehash *hash;	// memory to hash
	char fnhash[LEN_FILENAME];

	int nodes_end;		// current end position inside the file
	int nodes_fd;		// file descriptor
	char fnnodes[LEN_FILENAME];
};

struct s_osm_xmlcallback {
	FILE *f;
	BZFILE *bzf;
	long long int filesize;
	long long int filepos;
};

#define OSM_HASHNODES 16
#define OSM_HASHNODESMEM 1024

static struct s_osm_hashnodes _hashnodes[OSM_HASHNODES];
static struct s_osm_hashnodes *hashnodes[OSM_HASHNODES];

static int osm_mway_pmax = 0;
static struct map_way *osm_mway = NULL;
static int osm_marea_pmax = 0;
static struct map_area *osm_marea = NULL;

/* for statistic only */
unsigned long long int osm_cur_wayid;
struct map_pos osm_load_from_web_cur;
struct map_pos osm_load_from_web_start;
struct map_pos osm_load_from_web_end;
unsigned long long int osm_nodes_cnt;
unsigned long long int osm_area_cnt;
unsigned long long int osm_ways_cnt;
float osm_file_pos;

void osm_delete_tmpdir ();
void osm_init ();
void osm_free ();

char *osm_getfname_hid (char *fn, int size, short int hid, char *ext);
char *osm_getfname_id (char *fn, int size, unsigned long long int id, char *ext);

#define osm_hash_getid(_id) (short int)(_id/5000000)
struct s_osm_hashnodes *osm_hashnodes_load (short int hid, int loadhash);
void osm_hashnodes_free (struct s_osm_hashnodes *hnode);

#define osm_node_save(_node) _osm_node_save (_node, 0)
#define osm_node_append(_node) _osm_node_save (_node, 1)
int _osm_node_save (struct osm_node node, int append);
int osm_node_load (struct osm_node *node, unsigned long long int id);
int osm_hash_read (struct s_osm_hashnodes *hnodes, struct osm_nodehash *hash, int pos);
struct s_osm_hashnodes *osm_hash_get (unsigned long long int id, struct osm_nodehash *ret);

int osm_linehashsplit (struct map_pos p1, struct map_pos p2, struct map_pos *np);

struct osm_way *osm_way_alloc (int wmax);
void osm_way_copy (struct osm_way *d, struct osm_way *s);
void osm_way_save (struct osm_way *oway);
void osm_mway_new (int max);				/* realloc temp osm_way struct */

void osm_waynode_add (struct osm_node *node, int nodepos, unsigned long long int id,
                      unsigned short int subid, unsigned short int pnr);

void osm_marea_new (int max);
void osm_area_save (struct osm_way *oarea);

void osm_nodedetail_save (struct osm_nodedetail *onoded, struct osm_node *node);

/**************************************************************************
 * create a filename out of the node id..
 */
char *osm_getfname_id (char *fn, int size, unsigned long long int id, char *ext) {
	if (fn == NULL) return NULL;
	if (id == 0) return NULL;
	
	snprintf (fn, size, "%s/nodes-%d%s", TMPDIR, osm_hash_getid(id), ext);
	return fn;
};


char *osm_getfname_hid (char *fn, int size, short int hid, char *ext) {
	if (fn == NULL) return NULL;
	if (hid == -1) return NULL;
	
	snprintf (fn, size, "%s/nodes-%d%s", TMPDIR, hid, ext);
	return fn;
};


/**************************************************************************
 * do all preparations for loading a osm file
 */
void osm_delete_tmpdir (char *path) {
	DIR *dir;
	struct dirent *de;
	struct stat fstat;
	char fn[LEN_FILENAME];
	char fndir[LEN_FILENAME];
	static int _osm_recursive = 0;

	if (_osm_recursive > 16) {  // max number of calls..
		d_printf ("%s:%d too many subdirs abort deleting (%s).\n", __FILE__, __LINE__, path);
		return;
	}
	_osm_recursive++;

	/* read the directory */
	dir = opendir(path);
	while (dir != NULL && (de = readdir(dir)) != NULL) {
		snprintf (fndir, LEN_FILENAME, "%s/%s", path, de->d_name);

		/* directory? delete this dir */
		sprintf (fn, "%s/%s", path, de->d_name);
		stat (fn, &fstat);
		if (S_ISDIR (fstat.st_mode) && strcmp (de->d_name, ".") && strcmp (de->d_name, "..")) {
			osm_delete_tmpdir (fndir);
			rmdir (fndir);
		}
		
		/* file? delete file */
		else if (S_ISREG (fstat.st_mode)) {
			unlink (fndir);
		}
	}
	closedir (dir);
	_osm_recursive--;
};


void osm_init () {
	int i;
	
	osm_delete_tmpdir (TMPDIR);
	
	/* take care of the hash */
	for (i = 0; i < OSM_HASHNODES; i++) {
		_hashnodes[i].hid = -1;
		_hashnodes[i].hash = NULL;		
		_hashnodes[i].hash_alloc = 0;		
		_hashnodes[i].hash_cnt = 0;
		_hashnodes[i].hash_fd = 0;
		_hashnodes[i].nodes_end = 0;
		_hashnodes[i].nodes_fd = 0;
		hashnodes[i] = &_hashnodes[i];
	}

	if (osm_mway) ml_free (osm_mway);
	osm_mway = NULL;
	osm_mway_pmax = 0;
	osm_marea = NULL;
	osm_marea_pmax = 0;

	osm_nodes_cnt = 0;
	osm_area_cnt = 0;
	osm_ways_cnt = 0;

#ifdef _LINUX_
	mkdir (TMPDIR, 0755);
#else
	mkdir (TMPDIR);
#endif
};


void osm_free () {
	int i;

	/* take care of the hash files */
	do {
		for (i = 0; i < OSM_HASHNODES; i++) if (hashnodes[i]->hid != -1) {
			osm_hashnodes_free (hashnodes[i]);
			break;
		}
	}
	while (i < OSM_HASHNODES);
	
	if (osm_mway) ml_free (osm_mway);
	osm_mway = NULL;
	osm_mway_pmax = 0;

	if (osm_marea) ml_free (osm_marea);
	osm_marea = NULL;
	osm_marea_pmax = 0;

	osm_delete_tmpdir (TMPDIR);
};



/***********************************************************************
 * nodehash filemanagment
 */

/*
 * load the needed nodes bzw nodehash file into the memory
 * free unused memory if needed 
 */
struct s_osm_hashnodes *osm_hashnodes_load (short int hid, int loadhash) {
	struct stat fstathash, fstatnodes;
	int i, j, t;
	struct s_osm_hashnodes *hnodes = NULL;

	/* searching for the current hash inside the memory,
	 * when finished: j - first free entry, t - last used */
	for (j = -1, t = -1, i = 0 ; i < OSM_HASHNODES; i++) {
		if (hashnodes[i]->hid == hid) {
			hnodes = hashnodes[i];
			if (i > 0) for(j = i; j > 0; j--) hashnodes[j] = hashnodes[j-1];
			hashnodes[0] = hnodes;
			break;
		}

		else if (j == -1 && hashnodes[i]->hid == -1) j = i;
		else if (hashnodes[i]->hid != -1)
			t = i;
	}

	if (hnodes == NULL) {
		/* give us the first entry */
		if (t+1 == OSM_HASHNODES) osm_hashnodes_free (hashnodes[t]); // delete last one

		hnodes = hashnodes[OSM_HASHNODES-1];
		for(i = OSM_HASHNODES-1; i > 0; i--)
			hashnodes[i] = hashnodes[i-1];
	
		hashnodes[0] = hnodes;

		/* create the filenames for the files */
		osm_getfname_hid (hnodes->fnhash, LEN_FILENAME, hid, ".hashnode");
		osm_getfname_hid (hnodes->fnnodes, LEN_FILENAME, hid, ".nodes");
	
		/* check how much memory we will need to laod the data */
		if (stat (hnodes->fnhash, &fstathash) != 0)
			fstathash.st_size = 0;
		if (stat (hnodes->fnnodes, &fstatnodes) != 0)
			fstatnodes.st_size = 0;

		/* load the needed data */
		hnodes->hid = hid;
		hnodes->hash = NULL;
		hnodes->hash_fd = open (hnodes->fnhash, O_RDWR | O_CREAT, S_IWUSR | S_IRUSR);
		hnodes->hash_cnt = fstathash.st_size / sizeof (struct osm_nodehash);
		hnodes->nodes_fd = open (hnodes->fnnodes, O_RDWR | O_CREAT, S_IWUSR | S_IRUSR);
		hnodes->nodes_end = fstatnodes.st_size;
	}

	/*
	 * do we need to load the full hash data?
	 */
	if (loadhash && hnodes->hash == NULL) {
		hnodes->hash_alloc = 5000 + hnodes->hash_cnt;
		hnodes->hash = (struct osm_nodehash*) ml_malloc (hnodes->hash_alloc * sizeof (struct osm_nodehash));
		lseek (hnodes->hash_fd, 0, SEEK_SET);
		if (read (hnodes->hash_fd, hnodes->hash, hnodes->hash_cnt * sizeof (struct osm_nodehash)) < sizeof (struct osm_nodehash) * hnodes->hash_cnt) {
			d_printf ("%s:%d could not read data from file:%s", __FILE__, __LINE__, hnodes->fnhash);
			errorexit (-1);
		}
	}

	return hnodes;
};


/*
 * all data will be saved imedatly so we need only to close the file
 */
void osm_hashnodes_free (struct s_osm_hashnodes *hnode) {
	if (hnode->hid == -1) return;

	if ((hnode - _hashnodes) >= OSM_HASHNODES || (hnode - _hashnodes) < 0) {
		d_printf ("%s:%d something went very wrong: %d, max (OSM_HASHNODES): %d\n", __FILE__, __LINE__,
		          (hnode - _hashnodes), OSM_HASHNODES);
		errorexit (-1);
	}

	/* free all data */
	if (hnode->hash && hnode->hash_alloc > 0) {
		lseek (hnode->hash_fd, 0, SEEK_SET);
		write (hnode->hash_fd, hnode->hash, hnode->hash_cnt*sizeof(struct osm_nodehash));
		ml_free (hnode->hash);
		hnode->hash_alloc = 0;
		hnode->hash = NULL;
	}
	else if (hnode->hash) {
		ml_free (hnode->hash);
		hnode->hash_alloc = 0;
		hnode->hash = NULL;
	}		
	hnode->hash_cnt = 0;
	if (hnode->hash_fd > 0) {
		close (hnode->hash_fd);
		hnode->hash_fd = 0;
	}
	hnode->nodes_end = 0;
	if (hnode->nodes_fd > 0) {
		close (hnode->nodes_fd);
		hnode->nodes_fd = 0;
	}
	hnode->hid = -1;
};


/*
 * read hash node from position
 */
int osm_hash_read (struct s_osm_hashnodes *hnodes, struct osm_nodehash *hash, int pos) {
	if (hnodes->hash_alloc > 0 && hnodes->hash) {
		memcpy (hash, hnodes->hash+pos, sizeof(struct osm_nodehash));
	}
	else {
		lseek (hnodes->hash_fd, pos * sizeof (struct osm_nodehash), SEEK_SET);
		if (read (hnodes->hash_fd, hash, sizeof (struct osm_nodehash)) < sizeof (struct osm_nodehash)) {
			d_printf ("%s:%d could not read data from file:%s pos:%d", __FILE__, __LINE__, hnodes->fnhash, pos);
			return 0;
		}
	}
	return 1;
};


/* 
 * returns the nodehash entry to the givin node
 */
struct s_osm_hashnodes *osm_hash_get (unsigned long long int id, struct osm_nodehash *ret) {
	short int hid = osm_hash_getid (id);
	struct s_osm_hashnodes *hnodes = NULL;
	struct osm_nodehash hash;
	int left, pos, lpos1, lpos2, i;
	unsigned long long int lid;

	ret->id = 0;
	ret->pos = 0;

	/* prepare for looking inside the hash */
	hnodes = osm_hashnodes_load (hid, 0);
	left = hnodes->hash_cnt/2;
	pos = left;
	lpos2 = -2;
	lpos1 = -1;

	while (left > 0 && lpos1 != pos && lpos2 != pos && lpos1 != lpos2) {
		lpos1 = lpos2;
		lpos2 = pos;
		if (!osm_hash_read (hnodes, &hash, pos)) {
			d_printf ("%s:%d osm_hash_get id:%lld    hnodes.hid:%d hnodes.cnt:%d hnodes.fnnode:%s", __FILE__, __LINE__, id, hnodes->hid, hnodes->hash_cnt, hnodes->fnnodes);
			return NULL;
		}
		if (hash.id == id) {
			*ret = hash;
			return hnodes;
		}
		else if (hash.id < id) { 
			/* hid is larger than half.hid */
			left = left / 2 + left % 2;
			pos = pos+left;
			if (pos >= hnodes->hash_cnt) pos = hnodes->hash_cnt-1;
		}
		else {
			/* hid is smaler than half.hid */
			left = left / 2 + left % 2;
			pos = pos-left;
			if (pos < 0) pos = 0;
		}
	}

	if (!osm_hash_read (hnodes, &hash, pos)) {
		d_printf ("%s:%d osm_hash_get id:%lld    hnodes.hid:%d hnodes.cnt:%d hnodes.fnnode:%s", __FILE__, __LINE__, id, hnodes->hid, hnodes->hash_cnt, hnodes->fnnodes);
		return NULL;
	}
	if (hash.id == id) {
		*ret = hash;
		return hnodes;
	}
	else for (lpos1 = 0; lpos1 < left+1; lpos1++) {
		if (pos-lpos1 >= 0) {
			if (!osm_hash_read (hnodes, &hash, pos-lpos1)) {
				d_printf ("%s:%d osm_hash_get id:%lld    hnodes.hid:%d hnodes.cnt:%d hnodes.fnnode:%s", __FILE__, __LINE__, id, hnodes->hid, hnodes->hash_cnt, hnodes->fnnodes);
				return NULL;
			}
			if (hash.id == id) {
				*ret = hash;
				return hnodes;
			}
		}
		if (pos+lpos1 < hnodes->hash_cnt) {
			if (!osm_hash_read (hnodes, &hash, pos+lpos1)) {
				d_printf ("%s:%d osm_hash_get id:%lld    hnodes.hid:%d hnodes.cnt:%d hnodes.fnnode:%s", __FILE__, __LINE__, id, hnodes->hid, hnodes->hash_cnt, hnodes->fnnodes);
				return NULL;
			}
			if (hash.id == id) {
				*ret = hash;
				return hnodes;
			}
		}
	}

	d_printf ("node %lld not found over fast way. hid:%d", id, hnodes->hid);
	for (i = 0; i < OSM_HASHNODES; i++) {
		d_printf (" hid:%d hashcnt:%d hash_alloc:%d", hashnodes[i]->hid, hashnodes[i]->hash_alloc, hashnodes[i]->hash_cnt);
	}
	for (lid = 0, left = 0, pos = 1; (left < hnodes->hash_cnt); left++) {
		osm_hash_read (hnodes, &hash, left);
		if (hash.id == id) {
			d_printf ("     node found over slow way: %lld nr:%d", id, left);
			*ret = hash;
			return hnodes;
		}
		if (left>0 && pos) if (hash.id <= lid && hnodes->hash) {
			d_printf ("nodes are not sorted %lld <= %lld. hid:%d", hnodes->hash[left].id, hnodes->hash[left-1].id, hnodes->hid);
			pos = 0;
		}
		lid = hash.id;
	}
	
	return NULL;
};


/*
 * add the new hash entry to the hash table, we will not check for double
 * entrys, this have to be done by the caller.
 */ 
int osm_hash_add (struct s_osm_hashnodes *hnodes, struct osm_nodehash hash) {
	int i;
	struct osm_nodehash thash;

	if (hnodes->hash == NULL || hnodes->hash_alloc == 0)
		hnodes = osm_hashnodes_load (hnodes->hid, 1);
	if (hnodes->hash_alloc <= hnodes->hash_cnt) {
		hnodes->hash_alloc += 5000;
		hnodes->hash = ml_realloc (hnodes->hash, hnodes->hash_alloc * sizeof (struct osm_nodehash));
	}	

	/* check for the right place */
	i = hnodes->hash_cnt;
	if (i > 0) thash = hnodes->hash[i-1];
	while (i > 0 && thash.id > hash.id) {
		hnodes->hash[i] = thash;
		i--;
		if (i > 0) thash = hnodes->hash[i-1];
	}
	hnodes->hash[i] = hash;
	hnodes->hash_cnt++;
	
	return 1;
};


/***********************************************************************
 * loading / saving of a node also adding the new pos to the hash file
 */
int _osm_node_save (struct osm_node node, int append) {
	struct s_osm_hashnodes *hnodes = NULL;
	struct osm_nodehash hash;

	/* search if we have to update the node */
	hash.id = -1;
	hash.pos = -1;

	if (!append) /* update --> hash and hnodes == set */
		hnodes = osm_hash_get (node.id, &hash);

	/* append? check if we have enought memory also setup the hash entry */
	if (hnodes == NULL) { /* append */
		hnodes = osm_hashnodes_load (osm_hash_getid (node.id), 1);
		hash.pos = hnodes->nodes_end;
		hash.id = node.id;
		hnodes->nodes_end += sizeof (struct osm_node);
		osm_hash_add (hnodes, hash);
	}

	lseek (hnodes->nodes_fd, hash.pos, SEEK_SET);
	if (write (hnodes->nodes_fd, &node, sizeof (struct osm_node)) < sizeof (struct osm_node)) {
		d_printf ("%s:%d could not write data to file:%s pos:%d", __FILE__, __LINE__, hnodes->fnnodes, hash.pos);
		errorexit (-1);
	}

	return 1;
};


int osm_node_load (struct osm_node *node, unsigned long long int id) {
	struct s_osm_hashnodes *hnodes = NULL;
	struct osm_nodehash hash;

	hnodes = osm_hash_get (id, &hash);
	if (hnodes == NULL) return 0;

	lseek (hnodes->nodes_fd, hash.pos, SEEK_SET);
	if (read (hnodes->nodes_fd, node, sizeof (struct osm_node)) < sizeof (struct osm_node)) {
		d_printf ("%s:%d could not read data from file:%s pos:%d", __FILE__, __LINE__, hnodes->fnnodes, hash.pos);
		errorexit (-1);
	}
	return hash.pos;
};


void osm_waynode_add (struct osm_node *node, int nodepos, unsigned long long int id, 
                      unsigned short int subid, unsigned short int pnr) {
	struct s_osm_hashnodes *hnodes = NULL;
	struct osm_nodehash hash;
	struct osm_waynode wn;
	int pos, wpos;
	
	if (node == NULL) return;
	
	hnodes = osm_hash_get (node->id, &hash);
	if (hnodes == NULL) return;

	pos = hnodes->nodes_end;
	hnodes->nodes_end += sizeof (struct osm_waynode);

	/* save waynode at the end of the file */
	wn.wid = id;
	wn.wsubid = subid;
	wn.wpnr = pnr;
	wn.next = 0;
	lseek (hnodes->nodes_fd, pos, SEEK_SET);
	if (write (hnodes->nodes_fd, &wn, sizeof (struct osm_waynode)) < sizeof (struct osm_waynode)) {
		d_printf ("%s:%d could not write waynode data to file:%s pos:%d", __FILE__, __LINE__, hnodes->fnnodes, pos);
		errorexit (-1);
	}
						  
	if (node->waynode == 0) {
		/* save node with new waypos */
		node->waynode = pos;
		lseek (hnodes->nodes_fd, nodepos, SEEK_SET);
		if (write (hnodes->nodes_fd, node, sizeof (struct osm_node)) < sizeof (struct osm_node)) {
			d_printf ("%s:%d could not read data from file:%s pos:%d", __FILE__, __LINE__, hnodes->fnnodes, hash.pos);
			errorexit (-1);
		}
	}
	else {
		/* search for the last waynode entry */
		wpos = node->waynode;
		do {
			lseek (hnodes->nodes_fd, wpos, SEEK_SET);
			if (read (hnodes->nodes_fd, &wn, sizeof (struct osm_waynode)) < sizeof (struct osm_waynode)) {
				d_printf ("%s:%d could not read data from file:%s pos:%d", __FILE__, __LINE__, hnodes->fnnodes, hash.pos);
				errorexit (-1);
			}
			wpos = wn.next;
		}
		while (wn.next != 0);
		
		wn.next = pos;
		lseek (hnodes->nodes_fd, -sizeof (struct osm_waynode), SEEK_CUR);
		if (write (hnodes->nodes_fd, &wn, sizeof (struct osm_waynode)) < sizeof (struct osm_waynode)) {
			d_printf ("%s:%d could not write data to file:%s pos:%d", __FILE__, __LINE__, hnodes->fnnodes, hash.pos);
			errorexit (-1);
		}
	}

	return;
};


/****************************************************************************
 * split the line into peaces, is the line intersects with more then one
 * hash segment the function will return 2 as indicator of another 
 * intersection.
 */
int osm_linehashsplit (struct map_pos p1, struct map_pos p2, struct map_pos *np) {
	return 0; // for now just return 0 and get the other stuff running first.
};


/***************************************************************************
 * osm_way handling .. allocating memory and converting it into map_way
 */
struct osm_way *osm_way_alloc (int wmax) {
	struct osm_way *w = NULL;

	w = (struct osm_way*) ml_malloc (sizeof (struct osm_way) + wmax * sizeof (unsigned long long int));
	w->nmax = wmax;
	w->ncnt = 0;
	w->nid = (unsigned long long int*) &w->data[0];
	
	return w;
};


void osm_way_copy (struct osm_way *d, struct osm_way *s) {
	int i;
	
	if (d == NULL || s == NULL) {
		d_printf ("%s:%d  d or s is NULL \n", __FILE__, __LINE__);
		errorexit (-1);
	}
	
	d->wid = s->wid;
	d->type = s->type;
	strncpy (d->name, s->name, MAP_W_NAMELEN);
	for (i = 0; i < s->ncnt; i++)
		if (i < d->nmax) d->nid[i] = s->nid[i];
	d->ncnt = s->ncnt;
};


void osm_mway_new (int max) {
	if (osm_mway == NULL || max > osm_mway_pmax) {
		osm_mway = ml_realloc (osm_mway, sizeof (struct map_way) + 2 * sizeof (struct map_waynode) + max * sizeof (struct map_pos));
		osm_mway_pmax = max;
	}

	osm_mway->n_cnt = 0;
	osm_mway->p_cnt = 0;
	osm_mway->n = (struct map_waynode*) &osm_mway->data[0];
	osm_mway->p = (struct map_pos*) &osm_mway->data[2 * sizeof (struct map_waynode)];
	osm_mway->id = 0;
	osm_mway->subid = 0;
	osm_mway->next = 0;
};


/********************************************************************
 * convert the osm way into the map_way type and save it into the 
 * hash data block. I case the way goes over more hash segments the
 * way will be slit into peaces */
void osm_way_save (struct osm_way *oway) {
	int p;
	unsigned long long int id;
	struct osm_node node;
	struct map_pos curpos, lastpos;
	int nodepos;

	osm_mway_new (oway->ncnt);
	osm_mway->id = oway->wid;
	osm_mway->type = oway->type;
	strncpy (osm_mway->name, oway->name, MAP_W_NAMELEN);

	for (p = 0; p < oway->ncnt; p++) {
		/* find the node */
		id = oway->nid[p];
		if ((nodepos = osm_node_load (&node, id)) == -1)	return;

		/* check if we are still in the same AREA.. */
		curpos.lon = node.lon;
		curpos.lat = node.lat;
		
		if (p > 0 && 
			(map_geo2igeo (lastpos.lon) != map_geo2igeo (curpos.lon) || map_geo2igeo (lastpos.lat) != map_geo2igeo (curpos.lat))) {  
				 /* 
				  * save the known part of the way.. 
				  * and create a new part of the way
				  */
				 osm_mway->n[osm_mway->n_cnt].d_id = osm_mway->id;
				 osm_mway->n[osm_mway->n_cnt].d_subid = osm_mway->subid+1;
				 osm_mway->n[osm_mway->n_cnt].d_pnr = 0;
				 osm_mway->n[osm_mway->n_cnt].pnr = osm_mway->p_cnt-1;
				 osm_mway->n[osm_mway->n_cnt].d = curpos;
				 osm_mway->n_cnt = 1;
				 map_way_add (osm_mway, MHLOAD_RELOAD);
				
				 /* prepare a new way with subid + 1 */
				 osm_mway->n[osm_mway->n_cnt-1].d_subid = osm_mway->subid;
				 osm_mway->n[osm_mway->n_cnt-1].d_id = osm_mway->id;
				 osm_mway->n[osm_mway->n_cnt-1].d = lastpos;
				 osm_mway->n[osm_mway->n_cnt-1].d_pnr = osm_mway->p_cnt-1;
				 osm_mway->n[osm_mway->n_cnt-1].pnr = 0;
				 osm_mway->p_cnt = 0;
				 osm_mway->subid++;
			 }
		
		/*
		 * add the new node to the way.. also add the way to the nodehash..
		 */
		osm_mway->p[osm_mway->p_cnt].lon = node.lon;
		osm_mway->p[osm_mway->p_cnt].lat = node.lat;
		if (osm_mway->type < MWAY_MAX)
			osm_waynode_add (&node, nodepos, osm_mway->id, osm_mway->subid, osm_mway->p_cnt);
		else
			d_printf ("%s:%d osm_way_save: osm_mway->type >= MWAY_MAX which should not happen, since it should call osm_area_save", __FILE__, __LINE__);
		osm_mway->p_cnt++;
	
		/* save last position */
		lastpos = curpos;
		osm_cur_wayid = oway->wid;
		main_wnd_loop(p, -2);
	}

	 map_way_add (osm_mway, MHLOAD_RELOAD);
};


/******************************************************************************
 * all code to save an area
 */
void osm_marea_new (int max) {
	if (osm_marea == NULL || max > osm_marea_pmax) {
		osm_marea = ml_realloc (osm_marea, sizeof (struct map_area) + max * sizeof (struct map_pos));
		osm_marea_pmax = max;
	}

	osm_marea->id = 0;
	osm_marea->type = MAREA_unknown;
	osm_marea->p = (struct map_pos*) osm_marea->data;
	osm_marea->p_cnt = 0;
};


/********************************************************************
 * convert the osm "way" into the map_area type and save it into the 
 * hash data block. I case the way goes over more hash segments the
 * way will be slit into peaces 
 *
 * CALL with NULL will free all unneeded memory
 */
void osm_area_save (struct osm_way *oarea) {
	int ndfirst, nd;						// first node, current node
	int outfirstnd;							// first outside node, current outside node
	int i, j;
	static char *nodestat = NULL;			// array to save all some node data
	static int nodestat_alloc = 0;			// size of the struct
	struct map_pos pos, p1, outpos;			// current hash, pos
	struct map_hashpos curpos;
	struct map_pos lpos;
	struct osm_node curnode, node;			// current node, node

	/* check if we just need to free all data */
	if (oarea == NULL) {
		if (nodestat != NULL) free (nodestat);
		nodestat = NULL;
		nodestat_alloc = 0;
		return;
	}

	/* basic check and allocate enought memory (if needed) */
	if (oarea->ncnt < 2) return;	// a polygon needs at last 3 nodes
	osm_marea_new (oarea->ncnt);

	/* allocate data for nodestat, and reset data */
	if (nodestat_alloc < oarea->ncnt) {
		nodestat = ml_malloc (oarea->ncnt);
		nodestat_alloc = oarea->ncnt;
	}
	memset (nodestat, 0x0, nodestat_alloc);

	/* 
	 * find the first complete hash segment and it's end
	 */
	do {
		/* find first unused node */
		for (ndfirst = 0; ndfirst < oarea->ncnt && nodestat[ndfirst] != 0; ndfirst++);

		/* 
		 * found an unused node so save this segment 
		 */
		if (ndfirst < oarea->ncnt) {
			if (osm_node_load (&curnode, oarea->nid[ndfirst]) == -1) {
				d_printf ("%s:%d osm_area_save count not find node number: %lld", __FILE__, __LINE__, oarea->nid[ndfirst]);
				return;
			}
			curpos.ilon = map_geo2igeo (curnode.lon);
			curpos.ilat = map_geo2igeo (curnode.lat);
			nodestat[ndfirst] = 1;

			/* save first node */
			osm_marea->id = oarea->wid;
			osm_marea->type = oarea->type;
			osm_marea->p_cnt = 1;
			osm_marea->p = (struct map_pos *) osm_marea->data;
			osm_marea->p[0].lon = curnode.lon;
			osm_marea->p[0].lat = curnode.lat;

			/* find all nodes which are inside this segment
			 * do preparations after this save the area */
			nd = ndfirst;
			outfirstnd = -1;
			pos.lon = curnode.lon;
			pos.lat = curnode.lat;
			if (++nd >= oarea->ncnt) nd = 0;	/* count +1, check for overrun */
			while (nd != ndfirst) {
				lpos = pos;							/* save last node pos */
				
				if (osm_node_load (&node, oarea->nid[nd]) == -1) {
					d_printf ("%s:%d osm_area_save count not find node number: %lld", __FILE__, __LINE__, oarea->nid[nd]);
					return;
				}
				pos.lon = node.lon;
				pos.lat = node.lat;

				if (outfirstnd >= 0) {
					/* we are still outside, find the next inside pos */
					d_printf ("%s:%d FIX ME", __FILE__, __LINE__);
					exit (-1);
				}

				else {
					/* we are inside the current node
					 * check and split in two if needed */
					if (osm_linehashsplit (osm_marea->p[osm_marea->p_cnt-1], pos, &p1) > 0) {
						/* node is outside the area, the border is saved at p1 */
						outpos.lon = osm_marea->p[osm_marea->p_cnt].lon = p1.lon;
						outpos.lon = osm_marea->p[osm_marea->p_cnt].lat = p1.lat;
						osm_marea->p_cnt++;
						outfirstnd = nd;
					}
					else {
						/* node is inside the area */
						osm_marea->p[osm_marea->p_cnt].lon = node.lon;
						osm_marea->p[osm_marea->p_cnt].lat = node.lat;
						osm_marea->p_cnt++;
						nodestat[nd] = 1;
					}
				}
				if (++nd >= oarea->ncnt) nd = 0;	/* count +1, check for overrun */
			}
			map_area_add (osm_marea, MHLOAD_RELOAD);
		}
	} while (ndfirst < oarea->ncnt);
};



/********************************************************************
 * save detail information of the node. Thi will also create the 
 * search table.
 */
void osm_nodedetail_save (struct osm_nodedetail *onoded, struct osm_node *node) {
	int i, j, c;
	char *data;
	char *var;
	char *val;

	data = onoded->data;
	if (onoded->type == OSM_NODEDETAIL_PLACE) {
		struct map_place place;

		place.lon = node->lon;
		place.lat = node->lat;
		place.id = node->id;
		place.name = NULL;
		place.country = NULL;
		place.state = NULL;
		place.district = NULL;
		place.postalcode = NULL;
		place.next = NULL;
		place.population = 0;
		place.type = MAP_PLACE_unknown;

		for (i = 0; i < onoded->elements; i++) {
			var = data;
			val = var + strlen (var) + 1;
			data = val + strlen (val) + 1;

			// d_printf ("variable:%s   value:%s", var, val);
			if (strcmp (var, "name") == 0) {
				place.name = val;
			}
			else if (strcmp (var, "place") == 0) {
				for (j = 0; val[j] != 0; j++)
					val[j] = tolower (val[j]);
				if (strcmp (val, "city") == 0) {
					place.type = MAP_PLACE_city;
				}
				else if (strcmp (val, "suburb") == 0) {
					place.type = MAP_PLACE_suburb;
				}
				else if (strcmp (val, "village") == 0) {
					place.type = MAP_PLACE_village;
				}
				else if (strcmp (val, "hamlet") == 0) {
					place.type = MAP_PLACE_hamlet;
				}
				else if (strcmp (val, "town") == 0) {
					place.type = MAP_PLACE_town;
				}
				else {
					d_printf ("place:'%s' for '%s'", val, place.name);
					place.type = MAP_PLACE_unknown;
				}
			}
		}

		if (place.type != MAP_PLACE_unknown);
		map_place_add (&place, MHLOAD_RELOAD);
		map_search_add (place.country, place.name, place.id, place.lon, place.lat);
	}
	else if (onoded->type == OSM_NODEDETAIL_POI) {
	}
};


/*
 * append to the onoded another data, this will also check the size of the allocated data
 */
void osm_nodedetail_append (struct osm_nodedetail *onoded, char *variable, char *value) {
	char *data, *var, *val;
	int i,j;

	/* allocated memory for the data is too small */
	if (onoded->allocsize == 0 || onoded->data == NULL) {
		onoded->allocsize = strlen (variable) + strlen (value) + 2 + 4096;
		onoded->size = 0;
		onoded->data = ml_malloc (onoded->allocsize);
	}
	else if ((strlen (variable) + strlen (value) + 2 + onoded->size) >= onoded->allocsize) {
		char *olddata;
		int oldsize;

		olddata = onoded->data;
		oldsize = onoded->allocsize;
		onoded->allocsize += (strlen (variable) + strlen (value) + 2 + 4096);
		onoded->data = ml_malloc (onoded->allocsize);
		memcpy (onoded->data, olddata, oldsize);
		ml_free (olddata);
	}

	/* go to the last element and save data */
	data = onoded->data + onoded->size;
	var = data;
	for (i = 0; i < strlen (variable); i++)
		data[i] = tolower (variable[i]);
	data[i] = '\0';
	data = data + i + 1;
	val = data;
	for (j = 0; j < strlen (value); j++)
		data[j] = value[j];
	data[j] = '\0';
	onoded->size = onoded->size + j + i + 2;

	if (strncmp (var, "place", i) == 0) {
		onoded->type = OSM_NODEDETAIL_PLACE;
	}
	
	onoded->elements++;
};


/***************************************************************************
 * read a osmxml file, 
 */
#define XML_TEXTLEN 256
void osm_processdata (xmlTextReaderPtr reader) {
    const xmlChar *name, *attr, *k, *v;
	char *cid;
	unsigned long long int id;
	static struct osm_node onode;
	static struct osm_nodedetail onoded;
	static struct osm_way *oway = NULL;
	struct osm_way *tmpoway = NULL;
	static int olddepth = -1;
	static void *last = NULL;
	int depth;
	
	if (olddepth == -1) {
		onode.id = 0;
		onoded.data = NULL;
		onoded.allocsize = 0;
		onoded.size = 0;
		onoded.elements = 0;
		onoded.type = OSM_NODEDETAIL_NONE;
		if (oway != NULL) {
			oway->type = MWAY_unknown;
			oway->wid = 0;
			oway->ncnt = 0;
		}
		else {
			oway = osm_way_alloc (32);
		}
		last = NULL;
	}
	
    name = xmlTextReaderConstName(reader);
    if (name == NULL)
		name = BAD_CAST "--";
	
	/* check if we need to save the data.. */
	cid = (char*) xmlTextReaderGetAttribute (reader, (xmlChar*)"id");
	if (cid) {
		id = atoll (cid);
		free (cid);
	}
	else id = 0;
	depth = xmlTextReaderDepth (reader);
	
	if (depth == 1 && (strncmp ((char*) name, "#text", 6) == 0)) {
		/*
		 * save data
		 */
		if (last == (void*) oway) {
			if (oway->type >= MWAY_MAX) {
				oway->type -= MWAY_MAX;
				osm_area_cnt++;
				osm_area_save (oway);
			}
			else {
				osm_ways_cnt++;
				osm_way_save (oway);
			}
		}
		else if (last == (void*) &onode) {
			/* add the current node */
			osm_nodes_cnt++;
			osm_node_append (onode);
			if (onoded.type != OSM_NODEDETAIL_NONE)
				osm_nodedetail_save (&onoded, &onode);
		}
		onode.id = 0;
		onode.waynode = 0;
		onode.waynodecnt = 0;
		onoded.type = OSM_NODEDETAIL_NONE;
		onoded.size = 0;
		onoded.elements = 0;
		oway->wid = 0;
		oway->type = MWAY_unknown;
		oway->ncnt = 0;
		oway->name[0] = 0;

		last = NULL;
	}

	/*
	 * check and process date for a node
	 */
	if (strncmp ((char*) name, (char*) "node",5) == 0) { // it's a node
		if (id == 0) return;
		onode.id = id;
	
		// d_printf ("xml.. node: %lld\n", onode.id);
		attr = xmlTextReaderGetAttribute (reader, (xmlChar*)"lat");	// lat
		if (attr == NULL) return;
		sscanf ((char*)attr , "%g", &onode.lat);
		free ((void *)attr);
	
		attr = xmlTextReaderGetAttribute (reader, (xmlChar*)"lon");	// lon
		if (attr == NULL) return;
		sscanf ((char*)attr, "%g", &onode.lon);
		free ((void *)attr);
		
		olddepth = depth;
		last = (void*)&onode;
	}

	else if ((strncmp ((char*) name, (char*) "tag", 3) == 0) && last == (void*)&onode) { // tag to a node?
		k = xmlTextReaderGetAttribute (reader, (xmlChar*)"k");
		if (k == NULL) return;
		v = xmlTextReaderGetAttribute (reader, (xmlChar*)"v");
		if (v == NULL) {
			free ((void *)k);
			return;
		}

		onoded.id = onoded.id;
		osm_nodedetail_append (&onoded, (char*)k, (char*)v);
		free ((void *)k);
		free ((void *)v);
	}

	/*
	 * check and process data for way or an area
	 */
	else if (strncmp ((char*) name, (char*) "way",4) == 0) { // it's a way
		if (id == 0) return;
		oway->wid = id;
		
		olddepth = depth;
		last = (void*)oway;
	}
	
	else if ((strncmp ((char*) name, (char*) "tag", 3) == 0) && last == (void*)oway) { // tag to a way?
		k = xmlTextReaderGetAttribute (reader, (xmlChar*)"k");
		if (k == NULL) return;
		v = xmlTextReaderGetAttribute (reader, (xmlChar*)"v");
		if (v == NULL) {
			free ((void *)k);
			return;
		}

		// d_printf ("name:'%s'   val:'%s'", k, v);
		
		if (strncmp ((char*) k, (char*) "highway", 8) == 0) {
			if (strcmp ((char*) v, (char*) "primary") == 0) oway->type = MWAY_tertiary;
			else if (strcmp ((char*) v, (char*) "primary_link") == 0) oway->type = MWAY_tertiary;
			else if (strcmp ((char*) v, (char*) "tertiary") == 0) oway->type = MWAY_tertiary;
			else if (strcmp ((char*) v, (char*) "secondary") == 0) oway->type = MWAY_tertiary;
			else if (strcmp ((char*) v, (char*) "motorway") == 0) oway->type = MWAY_highway;
			else if (strcmp ((char*) v, (char*) "motorway_link") == 0) oway->type = MWAY_highway;
			else if (strcmp ((char*) v, (char*) "trunk") == 0) oway->type = MWAY_highway;
			else if (strcmp ((char*) v, (char*) "trunk_link") == 0) oway->type = MWAY_highway;
			else if (strcmp ((char*) v, (char*) "residential") == 0) oway->type = MWAY_residential;
			else if (strcmp ((char*) v, (char*) "living_street") == 0) oway->type = MWAY_residential;
			else if (strcmp ((char*) v, (char*) "road") == 0) oway->type = MWAY_residential;
			else if (strcmp ((char*) v, (char*) "unclassified") == 0) oway->type = MWAY_residential;
			else if (strcmp ((char*) v, (char*) "path") == 0) oway->type = MWAY_residential;
			else if (strcmp ((char*) v, (char*) "construction") == 0) oway->type = MWAY_residential;
			else if (strcmp ((char*) v, (char*) "cycleway") == 0) oway->type = MWAY_cycleway;
			else if (strcmp ((char*) v, (char*) "track") == 0) oway->type = MWAY_track;
			else if (strcmp ((char*) v, (char*) "service") == 0) oway->type = MWAY_service;
			else if (strcmp ((char*) v, (char*) "platform") == 0) oway->type = MWAY_footway;
			else if (strcmp ((char*) v, (char*) "footway") == 0) oway->type = MWAY_footway;
			else if (strcmp ((char*) v, (char*) "pedestrian") == 0) oway->type = MWAY_footway;
			else if (strcmp ((char*) v, (char*) "steps") == 0) oway->type = MWAY_footway;
			else {
				oway->type = MWAY_unknown;
				printf ("unknown highway:'%s'\n", v);
			}
		}

		/*
		 * check if it is a railway
		 */
		else if (strncmp ((char*) k, (char*) "railway", 8) == 0) {
			if (strcmp ((char*) v, (char*) "rail") == 0) oway->type = MWAY_track;
			else if (strcmp ((char*) v, (char*) "abandoned") == 0) oway->type = MWAY_track;
			else {
				oway->type = MWAY_unknown;
				d_printf ("unknown railway:'%s'", v);
			}
		}


		/*
		 * name ?
		 */
		else if (strncmp ((char*) k, (char*) "name", 5) == 0) {
			strncpy (oway->name, v, MAP_W_NAMELEN);
		}

		
		/*
		 * check if we have an area
		 */
		else if (strncmp ((char*) k, (char*) "natural", 8) == 0 || strncmp ((char*) k, (char*) "landuse", 8) == 0) {
			if (strcmp ((char*) v, (char*) "water") == 0) oway->type = MAREA_water;
			else if (strcmp ((char*) v, (char*) "industrial") == 0) oway->type = MAREA_industrial;
			else if (strcmp ((char*) v, (char*) "heath") == 0) oway->type = MAREA_industrial;
			else if (strcmp ((char*) v, (char*) "construction") == 0) oway->type = MAREA_industrial;
			else if (strcmp ((char*) v, (char*) "commercial") == 0) oway->type = MAREA_commercial;
			else if (strcmp ((char*) v, (char*) "retail") == 0) oway->type = MAREA_commercial;
			else if (strcmp ((char*) v, (char*) "cemetery") == 0) oway->type = MAREA_commercial;
			else if (strcmp ((char*) v, (char*) "farm") == 0) oway->type = MAREA_farm;
			else if (strcmp ((char*) v, (char*) "farmland") == 0) oway->type = MAREA_farm;
			else if (strcmp ((char*) v, (char*) "farmyard") == 0) oway->type = MAREA_farm;
			else if (strcmp ((char*) v, (char*) "village_green") == 0) oway->type = MAREA_greenfield;
			else if (strcmp ((char*) v, (char*) "greenfield") == 0) oway->type = MAREA_greenfield;
			else if (strcmp ((char*) v, (char*) "grass") == 0) oway->type = MAREA_greenfield;
			else if (strcmp ((char*) v, (char*) "residential") == 0) oway->type = MAREA_residential;
			else if (strcmp ((char*) v, (char*) "allotments") == 0) oway->type = MAREA_residential;
			else if (strcmp ((char*) v, (char*) "forest") == 0) oway->type = MAREA_wood;
			else if (strcmp ((char*) v, (char*) "meadow") == 0) oway->type = MAREA_wood;
			else if (strcmp ((char*) v, (char*) "wood") == 0) oway->type = MAREA_wood;
			else if (strcmp ((char*) v, (char*) "land") == 0) oway->type = MAREA_land;
			else if (strcmp ((char*) v, (char*) "military") == 0) oway->type = MAREA_military;
			else if (strcmp ((char*) v, (char*) "basin") == 0) oway->type = MAREA_water;
			else if (strcmp ((char*) v, (char*) "-----") == 0) oway->type = MAREA_unknown;
			else {
				oway->type = MAREA_unknown;
				d_printf ("unknown natural+landuse:'%s'", v);
			}
			oway->type += MWAY_MAX;
		}
		free ((void *)k);
		free ((void *)v);
	}
	
	else if ((strncmp ((char*) name, (char*) "nd", 3) == 0) && last == (void*)oway) { // node to a way?
		attr = xmlTextReaderGetAttribute (reader, (xmlChar*)"ref");
		if (attr == NULL) return;
		id = atoll ((char*)attr);
		free ((void *)attr);

		if (oway->ncnt+1 > oway->nmax) {
			tmpoway = osm_way_alloc (oway->nmax+16);
			osm_way_copy (tmpoway, oway);
			ml_free (oway);
			oway = tmpoway;
		}
		oway->nid[oway->ncnt] = id;
		oway->ncnt++;
	}
	else if (depth == 1) {
		olddepth = depth;
		last = NULL;
	}
};


int osm_loadfile_readcallback (void *context, char *buffer, int len) {
	int bzerr, bzlen, i;
	struct s_osm_xmlcallback *xmlcb = (struct s_osm_xmlcallback*) context;
	static int loopcnt;
	if (++loopcnt > 100) {
		xmlcb->filepos = ftell (xmlcb->f) / 1000;
		osm_file_pos = (100.0 * (float)(xmlcb->filepos) / (float)(xmlcb->filesize));
		loopcnt = 0;
	}
	
	bzlen = BZ2_bzRead (&bzerr, xmlcb->bzf, buffer, len);
	if (bzerr == BZ_STREAM_END && !feof(xmlcb->f)) {
		/* next data stream */
		unsigned char unused[BZ_MAX_UNUSED];
		unsigned char *unusedTmp;
		int nUnused = 0;

		BZ2_bzReadGetUnused (&bzerr, xmlcb->bzf, (void**) &unusedTmp, &nUnused);
		if (nUnused > 0) memcpy (unused, unusedTmp, nUnused);
		BZ2_bzReadClose (&bzerr, xmlcb->bzf);
		xmlcb->bzf = BZ2_bzReadOpen(&bzerr, xmlcb->f, 0, 0, unused, nUnused);
	}
	
	if (bzerr != BZ_OK) d_printf ("%s:%d bzread: should:%d is:%d   filepos:%d   error:%d", __FILE__, __LINE__, bzlen, len, xmlcb->filepos, bzerr);

	if (bzerr == BZ_CONFIG_ERROR) d_printf ("BZ_CONFIG_ERROR");
	else if (bzerr == BZ_PARAM_ERROR) d_printf ("BZ_PARAM_ERROR");
	else if (bzerr == BZ_IO_ERROR) d_printf ("BZ_IO_ERROR");
	else if (bzerr == BZ_MEM_ERROR) d_printf ("BZ_MEM_ERROR");
	else if (bzerr == BZ_SEQUENCE_ERROR) d_printf ("BZ_SEQUENCE_ERROR");
	else if (bzerr == BZ_DATA_ERROR) d_printf ("BZ_DATA_ERROR");
	else if (bzerr == BZ_DATA_ERROR_MAGIC) d_printf ("BZ_DATA_ERROR_MAGIC");
	else if (bzerr == BZ_UNEXPECTED_EOF) d_printf ("BZ_UNEXPECTED_EOF");
	else if (bzerr != BZ_OK) d_printf ("any other bz error is there:%d", bzerr);

	return bzlen;
};


int osm_loadfile_closecallback(void *context) {
    int bzerr;
    
    BZ2_bzReadClose(&bzerr, ((struct s_osm_xmlcallback *)context)->bzf);
    return bzerr == BZ_OK;
};


void osm_loadfile (char *fn) {
	xmlTextReaderPtr reader;
	struct s_osm_xmlcallback xmlcb;
	int ret, compressed, i, timeout_cnt, bzerr;
	time_t timeout;
	char ext[16];
	
	osm_init ();
	if (fn == NULL) return;
	main_wnd_status (_("converting file"), 0);
	d_printf ("osm_loadfile (%s)", fn);
	
	/* check the extension of the osm file */
	for (i = 0; i < 15 && fn[i+strlen(fn)-4] != 0; i++) 
		ext[i] = tolower (fn[i+strlen(fn)-4]);
	ext[i] = 0;
	if (strncmp (ext, ".bz2", 5) == 0) compressed = 1;
	else compressed = 0;

	/* set the locale to POSIX, needed for decimal numbers. */
	setlocale(LC_ALL, "C");

	if (compressed) {	    
		xmlcb.f = fopen (fn, "r");
		if (!xmlcb.f) {
			d_printf ("%s:%d could not open file '%s' for reading. Error:%s", __FILE__, __LINE__, fn, strerror (errno));
			return;
		}
		fseek (xmlcb.f, 0, SEEK_END);
		xmlcb.filesize = ftell (xmlcb.f) / 1000;
		fseek (xmlcb.f, 0, SEEK_SET);
	
		xmlcb.bzf = BZ2_bzReadOpen(&bzerr, xmlcb.f, 0, 0, NULL, 0);
	    if (xmlcb.bzf == NULL || bzerr != BZ_OK) {
			d_printf ("%s:%d osm_loadfile could not open source file.", __FILE__, __LINE__);
			exit (0);
	    }
		reader = xmlReaderForIO(osm_loadfile_readcallback, osm_loadfile_closecallback, &xmlcb, "", NULL, 0);
	}
	else reader = xmlReaderForFile(fn, NULL, 0);

	/*
	 * read the whole file...
	 */
	if (reader != NULL) {
		ret = xmlTextReaderRead(reader);
		while (ret == 1) {
			osm_processdata(reader);
			ret = xmlTextReaderRead(reader);
			
			if (++timeout_cnt > 1000 || timeout < time(NULL)) {	
				if (compressed)
					main_wnd_loop(xmlcb.filepos, xmlcb.filesize);
				else 
					main_wnd_loop(xmlTextReaderGetParserLineNumber (reader), -1);
				timeout = time(NULL);
				timeout_cnt = 0;
			}	
		}
		xmlFreeTextReader(reader);
	}
	if (compressed) {
		fclose (xmlcb.f);
	}

	/* reset the locale setting back to the environment setting */
	setlocale (LC_ALL, NULL);
	osm_free ();
	main_wnd_status (_(""), -1);
};


#define MLWEBMAX 10
void map_load_web (struct map_pos hs, struct map_pos he) {
	char parm1[128];
	char parm2[128];
	char app[255];
	char url[128];
	char *args[13];
	int ret, i;
	float x1, y1;
	struct stat sb;
	struct map_pos h;
	struct {
		struct s_execcall p;
		char fcache[LEN_FILENAME];
	} wl[MLWEBMAX];

#if defined (_LINUX_)
	app_status = APPSTATUS_loadfromweb;
	osm_load_from_web_end = he;
	osm_load_from_web_start = hs;
	main_wnd_status (_("refresh from OpenStreetMap.org"), 0);

	for (ret = 0; ret < MLWEBMAX; ret++) wl[ret].p.used = 0;
	for (ret = 0; ret < 13; ret++) args[ret] = NULL;

	for (h.lon = hs.lon; h.lon < he.lon || cmpf(h.lon, he.lon); h.lon = h.lon + MAP_OSMWEB_DELTA) 
		for (h.lat = hs.lat; h.lat < he.lat || cmpf(h.lat, he.lat); h.lat = h.lat + MAP_OSMWEB_DELTA) {
				do {
					main_wnd_loop (-1,-1);
					for (i = 0; i < MLWEBMAX; i++) {
						if (wl[i].p.used) {
							ret = execwait (&wl[i].p);
							if (ret == 1) {
								osm_loadfile (wl[i].fcache);
								draw ();
								main_wnd_status (_("refresh from OpenStreetMap.org"), 0);
								main_wnd_update ();
							}
						}
					}

					for (i = 0; i < MLWEBMAX && wl[i].p.used == 1; i++);
				}
				while (i >= MLWEBMAX);

				y1 = CALC_START (h.lon, MAP_OSMWEB_DELTA);
				x1 = CALC_START (h.lat, MAP_OSMWEB_DELTA);
			
				setlocale(LC_ALL, "C");
				sprintf (wl[i].fcache, "%s/c%.4f-%.4f-%.4f-%.4f.osm", cachepath, y1-MAP_OSMWEB_DELTA, x1-MAP_OSMWEB_DELTA, y1+MAP_OSMWEB_DELTA, x1+MAP_OSMWEB_DELTA);
				if (stat(wl[i].fcache, &sb) == -1) {
					setlocale(LC_ALL, "C");
					sprintf (app, "/usr/bin/wget");
					args[0] = app;
					sprintf (parm1, "-O");
					args[1] = parm1;
					sprintf (parm2, "%s", wl[i].fcache);
					args[2] = parm2;
					sprintf (url, "http://api.openstreetmap.org/api/0.6/map?bbox=%.5f,%.5f,%.5f,%.5f", y1-MAP_OSMWEB_DELTA, x1-MAP_OSMWEB_DELTA, y1+MAP_OSMWEB_DELTA, x1+MAP_OSMWEB_DELTA);
					args[3] = url;
					setlocale(LC_ALL, NULL);
					ret = execcall(&wl[i].p, app, args);
				}
				else {
					osm_loadfile (wl[i].fcache);
					main_wnd_update ();
				}
			}
	do {
		main_wnd_loop (-1,-1);
		for (i = 0; i < MLWEBMAX; i++) {
			if (wl[i].p.used) {
				ret = execwait (&wl[i].p);
				if (ret == 1) {
					osm_loadfile (wl[i].fcache);
					main_wnd_update ();
				}
			}
		}

		for (ret = 0, i = 0; i < MLWEBMAX; i++) 
			if (wl[i].p.used == 1) ret = 1;
		}
	while (ret);
#endif
	app_status = APPSTATUS_nothing;
};


/***************************************************
 * displaying some informations on the screen 
 */
#ifdef COSMROUTE
#include <curses.h>

void osm_convert_info () {
	int i;
	
	attrset(A_BOLD | COLOR_PAIR(3));
	mvprintw(6,10, "OSM HashNodes  Information");
	attrset(COLOR_PAIR(3));
	for (i = 0; i < 16; i++)
			mvprintw(7+i,1, "hid:     hash:          nodes:       kb");
	attrset(A_BOLD | COLOR_PAIR(2));
	for (i = 0; i < 16; i++) {
		mvprintw(7+i,5, "%d", hashnodes[i]->hid);
		mvprintw(7+i,15, "%d", hashnodes[i]->hash_cnt);
		mvprintw(7+i,31, "%d", hashnodes[i]->nodes_end/1024);
		if (hashnodes[i]->hash) mvprintw(7+i, 0, "*");
		else mvprintw(7+i, 0, " ");
	}
}

#endif
