/***************************************************************************
 *            map_way.c
 *
 *  Copyright  2009 - Steffen Pohle
 *  steffen@gulpe.de
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "map.h"
#include "memoryleak.h"
#include "system.h"
#include "vector.h"



/*****************************************************
 *
 *  Way handling
 *
 */

/* 
 * add the way without checking for it's size, but do an update
 */
int map_way_add_to_hash (struct map_hash *mh, struct map_way *mw) {
	int size;
	struct map_way *w;

	/* check if the first element is already the one we want to update */
	if (mh->ways && mh->ways->id == mw->id && mh->ways->subid == mw->subid) {
		mh->ways = mh->ways->next;
		mh->wayscnt--;
	}

	/* find possible update and go to the end */
	if (mh->ways) {
		/* find last way */	
		w = mh->ways;
		while (w && w->next) {
			if (w->next->id == mw->id && w->next->subid == mw->subid) {
				/* found old way.. just delete .. */
				w->next = w->next->next;
				mh->wayscnt--;
			}
			if (w->next) w = w->next;
		}
		w->next = (struct map_way*) map_hash_getfree_aligned (mh, POINTERALIGNMENT);
		w = w->next;
	}
	else {
		/* complete new ways */
		mh->ways = (struct map_way*) map_hash_getfree_aligned (mh, POINTERALIGNMENT);
		w = (struct map_way*) map_hash_getfree_aligned (mh, POINTERALIGNMENT);
	}
	
	/* copy the way data to the free entry
	 * calc size and set free entry to it's new place */
	size = map_way_getsize (mw);
	map_way_copy (w, mw);
	w->next = NULL;
	mh->free = mh->free + size;
	mh->wayscnt++;
	SETFLAG(mh->flags, MHF_changed);
	mh->tickstamp = getticks();

	return 1;
};


/******************************************************************************
 * Add/Update way information
 * - load/ get hash table
 * - calculate size in bytes for this entry
 * - check if pointer in free is enought for this new entry
 *   if not: move old memory block out of the way and put a new data block inside.
 * - add new way
 */
int map_way_add (struct map_way *way, int loadflags) {
	struct map_hash *mh = NULL;
	int size, sway, eway;

	if (way->p == NULL || way->p_cnt == 0) return 0;
	size = map_way_getsize (way);
	mh = map_hash_get (way->p[0].lon, way->p[0].lat, loadflags);
	sway = map_hash_countways (mh);

	if (mh == NULL) {
		mh = map_hash_alloc (size + MAP_HASH_DEFAULTSIZE);
		mh->pos.ilon = map_geo2igeo (way->p[0].lon);
		mh->pos.ilat = map_geo2igeo (way->p[0].lat);
		map_hash_add (mh);
	}
	else if (size >= map_hash_getfree (mh)) {
		mh = map_hash_realloc (mh, mh->datasize + size + MAP_HASH_DEFAULTSIZE);
	}
	
	return map_way_add_to_hash (mh, way);
};


/*
 * will copy all the data from one way to another, this function will not
 * check if there arew conflicts with the memory at all. Also the next field
 * will be keeped untouched.
 */
void map_way_copy (struct map_way *dest, struct map_way *src) {
	int i;

	strncpy (dest->name, src->name, MAP_W_NAMELEN);
	dest->id = src->id;
	dest->subid = src->subid;
	dest->type = src->type;
	dest->p_cnt = src->p_cnt;
	dest->n_cnt = src->n_cnt;
	dest->p = (struct map_pos*) (dest->data);
	dest->n = (struct map_waynode*) (dest->data+(sizeof (struct map_pos)*dest->p_cnt));
	for (i = 0; (i < dest->p_cnt || i < dest->n_cnt); i++) {
		if (i < dest->p_cnt) dest->p[i] = src->p[i];
		if (i < dest->n_cnt) dest->n[i] = src->n[i];
	}
};


struct map_way *map_way_find (struct map_hash *mh, unsigned long long int id, unsigned short int subid) {
	struct map_way *mw;

	mw = mh->ways;
	while (mw) {
		if (mw->id == id && subid == mw->subid) return mw;
		mw = mw->next;
	}

	return NULL;
};


int map_way_getsize (struct map_way *way) {
	return (sizeof (struct map_way) + way->n_cnt * sizeof (struct map_waynode) + way->p_cnt * sizeof (struct map_pos));
};


/************************************************************
 * find closest way to the givin position
 * idh: distance to check (igeo distance)
 */
unsigned long long int map_way_findclosest (struct map_pos *pos, struct map_pos *waypos, int idh, char *name, int namelen) {
	struct map_hash *mh;
	struct map_hashpos ipos;
	struct map_way *mw;
	struct map_pos tmppos;
	fPoint v1, v2, v3, v4;
	fPoint p1, p2, p3, p4;
	float dist_closest = 40000.0, f;
	unsigned long long int closest = 0;
	int i;

#if defined(COSMROUTE) || defined(OSM2MAP)
	float view_lon = pos->lon;
	float view_lat = pos->lat;
#endif

	p3.x = map_lon2km (view_lon - pos->lon, pos->lat);
	p3.y = map_lat2km (view_lat - pos->lat);

	ipos.ilon = map_geo2igeo (pos->lon);
	ipos.ilat = map_geo2igeo (pos->lat);

	mh = map_hash_geti (ipos.ilon, ipos.ilat, MHLOAD_RELOAD);
	if (mh == NULL) return 0;

	mw = mh->ways;
	while (mw) {
		for (i = 1; i < mw->p_cnt; i++) {
			/* mw->p... must be converted to km first. */
			p1.x = map_lon2km (view_lon - mw->p[i].lon, mw->p[i].lat);
			p1.y = map_lat2km (view_lat - mw->p[i].lat);
			p2.x = map_lon2km (view_lon - mw->p[i-1].lon, mw->p[i-1].lat);
			p2.y = map_lat2km (view_lat - mw->p[i-1].lat);

			v1 = vec_sub (p1, p2);
			v2 = vec_sub (p3, p2);
			v3 = vec_normale (v1, v2);

			/* need to check the range f the new vec.. */
			if     ((v3.x <= 0.0 && v1.x <= 0.0 || v3.x >= 0.0 && v1.x >= 0.0 
			         || v3.y <= 0.0 && v1.y <= 0.0 || v3.y >= 0.0 && v1.y >= 0.0)
			        && fabs(v3.x) <= fabs(v1.x) && fabs(v3.y) <= fabs(v1.y)) {
				
				v4 = vec_sub (v3, v2);
				f = sqrt (v4.x * v4.x + v4.y * v4.y);
				if (f < dist_closest) {
					dist_closest = f;
					closest = mw->id;
					p4 = vec_add (v3, p2);
					tmppos.lat = view_lat - map_km2lat (p4.y);
					tmppos.lon = view_lon - map_km2lon (p4.x, tmppos.lat);
					*waypos = tmppos;
					strncpy (name, mw->name, namelen);
				}
			}
		}
		mw = mw->next;
	};
	
	return closest;
};

