/*
 * memoryleak.c
 * Copyright (C) Steffen Pohle 2009 <steffen@gulpe.de>
 * 
 * main.c is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * main.c is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "memoryleak.h"

#if !defined(__MINGW32CE__) && !defined(_WIN32_WCE) && !defined(__MINGW32__)
#define MEMORY_DUMPINFO 1000

struct memoryleak_data memory[MEMORYLEAK_MAX];
int memory_used = 0;
int memory_size = 0;
int memory_blocknext = 0;

static struct memoryleak_data *ml_findfree () {
	int i;

	for (i = 0; i < MEMORYLEAK_MAX; i++)
		if (memory[i].ptr == NULL) return &memory[i];

	return NULL;
};


static struct memoryleak_data *ml_find (void *ptr) {
	int i;

	for (i = 0; i < MEMORYLEAK_MAX; i++)
		if (memory[i].ptr == ptr) return &memory[i];
	
	return NULL;
};


void *ml_calloc(size_t nmemb, size_t size) {
	return calloc(nmemb, size);
};


void *ml_malloc(size_t size) {
	struct memoryleak_data *md = ml_findfree ();
    int i, j, nptrs;
    void *buffer[ML_BT_SIZE];
	char **strings = NULL;

	if (md == NULL) {
		d_printf ("memoryleak malloc MEMORYLEAK_MAX reached\n");

		for (j = 0; j < MEMORYLEAK_MAX; j++) {
			d_printf ("memoryleak : (%6d) size:%d  ptr:%p  text:%s", j, memory[j].size, memory[j].ptr, memory[j].text);
			for (i = 1; i < memory[j].btcnt; i++) if (memory[j].ptr) // don't show ml_malloc call.
				d_printf ("    frame:%3d %s", i, memory[j].btstr[i]);
		}
		errorexit (-1);
	}

	if (size > MEMORYLEAK_MAXMEM) {
		d_printf ("ml_realloc: too much memory %d", size);
		errorexit (-1);
	}
	
	md->ptr = malloc (size);
	md->size = size;
	md->btcnt = backtrace(buffer, ML_BT_SIZE);

	/* if someone finds out how this should really work please let me know? */
//	strings = backtrace_symbols(buffer, md->btcnt);
//    if (strings == NULL) {
//		perror("backtrace_symbols");
//		errorexit(-1);
//	}
	for (i = 0; (i < md->btcnt) && (i < ML_BT_SIZE); i++) {
//		d_printf ("i:%d  strings:%p", i, strings);
		snprintf (md->btstr[i], ML_BT_TEXT, "%p", buffer[i]);
	}
//	free (strings);

	md->text[0] = 0;
	
	memory_used ++;
	memory_size += size;

	return md->ptr;
};


void ml_free(void *ptr) {
	struct memoryleak_data *md = ml_find (ptr);

	if (md == NULL) {
		d_printf ("memoryleak free NULL\n");
		errorexit (-1);
	}

	memory_used--;
	memory_size -= md->size;

	free (md->ptr);
	md->ptr = NULL;
	md->text[0] = 0;
	md->size = 0;

};


void *ml_realloc(void *ptr, size_t size) {
	struct memoryleak_data *md = ml_find (ptr);
	void *test = NULL;
	
	if (ptr == NULL) md = ml_findfree ();
	if (md == NULL) {
		d_printf ("memoryleak realloc NULL\n");
		errorexit (-1);
	}

	if (size > MEMORYLEAK_MAXMEM) {
		d_printf ("ml_realloc: too much memory %d", size);
		errorexit (-1);
	}

	memory_size = memory_size - md->size + size;
	
	test = realloc (ptr, size);
	md->ptr = test;
	md->size = size;
	
	return md->ptr;
};


void ml_addinfo (void *ptr, char *text) {
	struct memoryleak_data *md = ml_find (ptr);

	if (md) {
		strncpy (md->text, text, ML_BT_TEXT);
	}
};


void ml_get_stats (struct memoryleak_info *m) {
	m->blocks = memory_used;
	m->size = memory_size;

//	if (memory_used > memory_blocknext) {
//		ml_debug ();
//		while (memory_used > memory_blocknext) memory_blocknext += MEMORY_DUMPINFO;
//	}
};


void ml_init () {
	int i;
	
	for (i = 0; i < MEMORYLEAK_MAX; i++) {
		memory[i].size = 0;
		memory[i].ptr = NULL;
	}	
};


void ml_debug () {
	int i, j;
	d_printf ("*** memoryleak debug:");
	for (j = 0; j < MEMORYLEAK_MAX; j++) if (memory[j].ptr) {
		d_printf (" memory(%6d) size:%d  ptr:%p   text:%s", j, memory[j].size, memory[j].ptr, memory[j].text);
		for (i = 1; i < memory[j].btcnt; i++) if (memory[j].ptr) // don't show ml_malloc call.
			d_printf ("    frame:%3d %s", i, memory[j].btstr[i]);
	}
};

#else

void ml_get_stats (struct memoryleak_info *m) {
};

void ml_debug () {
};

void ml_addinfo (void *ptr, char *text) {
};

#endif
