/***************************************************************************
 *            system.c
 *
 *  Copyright  2009 - Steffen Pohle
 *  steffen@gulpe.de
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "system.h"
#include "map.h"
#include "osmroute.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>


/*******************************************
 *
 * comparsion between doubles and floats 
 */
int cmpf (float f1, float f2) {
	return fabs(f1-f2) <= FLT_EPSILON;
};

int cmpd (double d1, double d2) {
	return fabs(d1-d2) < DBL_EPSILON;
};


#if !defined(__MINGW32CE__) && !defined(_WIN32_WCE) && !defined(__MINGW32__)
int execcall (struct s_execcall *p, char *prg, char **argv) {
	if (p->used) return 0;
	
	if ((p->pid = fork ()) < 0) {
		printf ("%s:%d syscall fork error: %s\n", __FILE__, __LINE__, strerror (errno));
		return 0;
	}
	else if (p->pid == 0) { // child //
		if (execvp(prg, argv) < 0) {
			printf ("%s:%d syscall execvp error:%s\n", __FILE__, __LINE__, strerror (errno));
			errorexit (-1);
		}
	}
	/* parent */
	
	p->used = 1;
	return 1;
};


int execwait (struct s_execcall *p) {
	pid_t ptmp;
	int status;
	
	if (p->used == 0) return -1;
	ptmp = waitpid (p->pid, &status, WNOHANG);
	if (ptmp == p->pid) {
		if (WIFEXITED (status)) {
			p->used = 0;
			return 1;
		}
	}
	else if (ptmp == -1) {
		p->used = 0;
		return 1;
	}

	return -2;
};
#endif



void d_printf (char *fmt,...) {
 	va_list args;
	char text[1024];
	char fn[LEN_FILENAME];
	int f;

	va_start (args, fmt);
 	vsnprintf (text, 1024, fmt, args);
 	va_end (args);
	text[1023] = 0;

	snprintf (fn, LEN_FILENAME, "%slogfile.txt", logpath);
	
#ifndef COSMROUTE
	fprintf (stderr, "%s\n", text);
#endif

	f = open (fn, O_RDWR | O_CREAT, 0644);
	if (f) {
		lseek (f, 0, SEEK_END);
		write (f, text, strlen (text));
#if !defined(__MINGW32CE__) && !defined(_WIN32_WCE) && !defined(__MINGW32__)
		write (f, "\n", 1);
#else
		write (f, "\n\r", 2);
#endif
		close (f);
	}
};


void d_print_init() {
	int i;
	char **strings;
	char fn[LEN_FILENAME];
	
	snprintf (fn, LEN_FILENAME, "%slogfile.txt", logpath);
	unlink (fn);
	d_printf ("");
	d_printf ("************************************************************");
	d_printf ("*               Starting                                   *");
	d_printf ("************************************************************");
	d_printf ("size of datatypes in bytes - should be on all systems the same");
	d_printf ("     type       : size ");
	d_printf ("----------------+------");
	d_printf ("float           : %d", sizeof (float));
	d_printf ("char            : %d", sizeof (char));
	d_printf ("uint16_t        : %d", sizeof (uint16_t));
	d_printf ("uint32_t        : %d", sizeof (uint32_t));
	d_printf ("uint64_t        : %d", sizeof (uint64_t));
	d_printf ("size of datatypes in bytes - can be different");
	d_printf ("     type       : size ");
	d_printf ("----------------+------");
	d_printf ("long long int   : %d", sizeof (long long int));
	d_printf ("long int        : %d", sizeof (long int));
	d_printf ("int             : %d  INT_MIN:%d, INT_MAX:%d", sizeof (int), INT_MIN, INT_MAX);
	d_printf ("short int       : %d", sizeof (short int));
	d_printf ("\nTest for Backtraces:");

	strings = get_backtrace (&i);
	if (i < 1 || strings == NULL) d_printf ("seems not to work\n");
	else {
		d_printf ("works: %s\n", strings[2]);
		free (strings);
	}
};


unsigned long long int getticks () {
	static unsigned long long int ticks = 0;

	if (ticks > (unsigned long long int)(ticks+1)) {
		d_printf ("%s:%d ticks+1 < ticks", __FILE__, __LINE__);
		errorexit (-1);
	}

	return ++ticks;
};


#define SIZE 100
void errorexit (int nr) {
	static int calledonce = 0;
	int j, nptrs;
	void *buffer[SIZE];
	char **strings;

#if !defined(__MINGW32CE__) && !defined(_WIN32_WCE) && !defined(__MINGW32__)
	nptrs = backtrace(buffer, SIZE);
	d_printf("****************** backtrace:  (needs to be linked with -Wl,--export-dynamic)");

	/* The call backtrace_symbols_fd(buffer, nptrs, STDOUT_FILENO)
	 * would produce similar output to the following: */
	strings = (char **) backtrace_symbols(buffer, nptrs);
	if (strings == NULL) {
		d_printf ("backtrace_symbols: %s", strerror(errno));
		exit(EXIT_FAILURE);
	}

	for (j = 0; j < nptrs; j++)
		d_printf("%s", strings[j]);

	free(strings);

#endif
	
	if (calledonce) exit (nr);
	calledonce = 1;
	map_save_all ();
	
	exit (nr);
};


void d_print_backtrace () {
	int j, nptrs;
	void *buffer[SIZE];
	char **strings;

#if !defined(__MINGW32CE__) && !defined(_WIN32_WCE) && !defined(__MINGW32__)
	nptrs = backtrace(buffer, SIZE);
	d_printf("****************** backtrace:  (needs to be linked with -Wl,--export-dynamic)");

	/* The call backtrace_symbols_fd(buffer, nptrs, STDOUT_FILENO)
	 * would produce similar output to the following: */
	strings = (char **) backtrace_symbols(buffer, nptrs);
	if (strings == NULL) {
		d_printf ("  backtrace_symbols: %s", strerror(errno));
		exit(EXIT_FAILURE);
	}

	for (j = 0; j < nptrs; j++)
		d_printf("  %s", strings[j]);

	free(strings);
#endif
};



char **get_backtrace (int *cnt) {
#if !defined(__MINGW32CE__) && !defined(_WIN32_WCE) && !defined(__MINGW32__)
	void *buffer[SIZE];
	char **strings;

	*cnt = backtrace(buffer, SIZE);

	/* The call backtrace_symbols_fd(buffer, nptrs, STDOUT_FILENO)
	 * would produce similar output to the following: */
	strings = (char **) backtrace_symbols(buffer, *cnt);
	if (strings == NULL) {
		d_printf ("backtrace_symbols: %s", strerror(errno));
		exit(EXIT_FAILURE);
	}

	return strings;
#else
	return NULL;
#endif
};


/*
 * copy data to the givin address and return increased pointer
 */
char *mempush (void *mem, void *source, int size) {
	memcpy (mem, source, size);
	return mem+size;
};

char *mempop (void *mem, void *dest, int size) {
	memcpy (dest, mem, size);
	return mem+size;
};

char *memstrpop (void *mem, void *dest, int *size) {
	uint8_t *m = mem, *d = dest;
	void *mout;
	
	while (*m != 0) {
		*d = *m;
		d++;
		m++;
	} 
	*d = *m;
	m++;

	mout = m;
	*size = mout - mem;

	return m;
};


struct stat *file_exist (char *fn) {
	static struct stat fs;

	if (stat(fn, &fs) != 0)
		return NULL;
	
	return &fs;
};
